"""Provides functions for manipulating molecules."""

from __future__ import annotations

from copy import deepcopy
from logging import getLogger

import numpy as np
from rdkit import RDLogger
from rdkit.Chem.AllChem import AssignBondOrdersFromTemplate
from rdkit.Chem.Lipinski import HAcceptorSmarts, HDonorSmarts
from rdkit.Chem.rdchem import AtomValenceException, Bond, Conformer, GetPeriodicTable, Mol, RWMol
from rdkit.Chem.rdForceFieldHelpers import UFFGetMoleculeForceField
from rdkit.Chem.rdMolAlign import GetBestAlignmentTransform
from rdkit.Chem.rdmolfiles import MolFromSmarts
from rdkit.Chem.rdmolops import AddHs, RemoveHs, RemoveStereochemistry, RenumberAtoms, SanitizeMol
from rdkit.Chem.rdMolTransforms import TransformConformer

logger = getLogger(__name__)


def copy_pos_to_template_mol(mol_ref: Mol, mol_pos: Mol) -> Mol:
    """Copy positions from mol_pos to mol_ref if possible."""
    mol_new = deepcopy(mol_ref)
    mol_new.RemoveAllConformers()
    RemoveStereochemistry(mol_new)

    RDLogger.DisableLog("rdApp.")  # type: ignore[attr-defined]
    mol_pos = AssignBondOrdersFromTemplate(refmol=mol_new, mol=mol_pos)
    RDLogger.EnableLog("rdApp.")  # type: ignore[attr-defined]
    assert mol_pos is not None

    # get atom mapping between molecules
    matches = mol_pos.GetSubstructMatches(mol_new, uniquify=True)
    assert len(matches) >= 1, "No matches found between molecules."

    # get positions and reorder to match atom mapping
    assert mol_pos.GetNumConformers() >= 1, "No conformers found in mol_pos."
    pos_raw = mol_pos.GetConformer().GetPositions()
    pos_new = pos_raw[matches[0], :]

    # create new conformer
    conformer = Conformer(mol_new.GetNumAtoms())
    for i, pos in enumerate(pos_new):
        conformer.SetAtomPosition(i, pos)

    # add conformer to mol
    mol_new.AddConformer(conformer)
    assert mol_new.GetNumConformers() == 1

    SanitizeMol(mol_new)
    return mol_new


def neutralize_atoms(mol: Mol) -> Mol:
    """Add and remove hydrogens to neutralize charges ignoring overall charge."""
    # https://www.rdkit.org/docs/Cookbook.html#neutralizing-charged-molecules
    # stronger than rdkit.Chem.MolStandardize.rdMolStandardize.Uncharger
    try:
        pattern = MolFromSmarts("[+1!h0!$([*]~[-1,-2,-3,-4]),-1!$([*]~[+1,+2,+3,+4])]")
        at_matches = mol.GetSubstructMatches(pattern)
        at_matches_list = [y[0] for y in at_matches]
        if len(at_matches_list) > 0:
            for at_idx in at_matches_list:
                atom = mol.GetAtomWithIdx(at_idx)
                chg = atom.GetFormalCharge()
                hcount = atom.GetTotalNumHs()
                atom.SetFormalCharge(0)
                atom.SetNumExplicitHs(hcount - chg)
                atom.UpdatePropertyCache()
    except AtomValenceException:
        logger.warning("AtomValenceException raised while neutralizing molecule. Continuing with original molecule.")
        return mol
    return mol


def remove_all_charges_and_hydrogens(mol: Mol) -> Mol:
    """Remove all charges and explicit hydrogens from molecule."""
    try:
        # rdkit keeps hydrogens that define sterochemistry so remove stereo first
        RemoveStereochemistry(mol)
        mol = RemoveHs(mol)
        # remove charges
        mol = neutralize_atoms(mol)
    except AtomValenceException:
        # from rdkit.Chem.MolStandardize.rdMolStandardize import Uncharger
        # mol = Uncharger().uncharge(mol)
        logger.warning("AtomValenceException raised while neutralizing molecule. Continuing with original molecule.")
        return mol
    return mol


def add_stereo_hydrogens(mol: Mol) -> Mol:
    """Add all hydrogens but those on primary ketimine."""
    exclude = {match[1] for match in mol.GetSubstructMatches(MolFromSmarts("[CX3]=[NH1]"))}
    atoms = [a.GetIdx() for a in mol.GetAtoms() if a.GetAtomicNum() != 1 if a.GetIdx() not in exclude]
    mol = AddHs(mol, onlyOnAtoms=atoms, addCoords=True)
    return mol


def remove_isotopic_info(mol: Mol) -> Mol:
    """Remove isotopic information from molecule."""
    for atom in mol.GetAtoms():
        atom.SetIsotope(0)
    return mol


def get_hbond_acceptors(mol: Mol) -> set[int]:
    """Return indices of atoms that can act as hydrogen bond acceptors."""
    return {s[0] for s in mol.GetSubstructMatches(HAcceptorSmarts, uniquify=1)}  # type: ignore[call-overload]


def get_hbond_donors(mol: Mol) -> set[int]:
    """Return indices of atoms that can act as hydrogen bond donors."""
    return {s[0] for s in mol.GetSubstructMatches(HDonorSmarts, uniquify=1)}  # type: ignore[call-overload]


def delete_atoms(mol: Mol, indices: list[int]) -> Mol:
    """Delete atoms from molecule.

    Args:
        mol: Molecule to delete atoms from.

    Returns:
        Molecule without atoms.
    """
    # delete in reverse order to avoid reindexing issues
    indices = sorted(indices, reverse=True)
    if len(indices) == 0:
        return mol
    mol = RWMol(mol)
    for index in indices:
        mol.RemoveAtom(index)
    return Mol(mol)


def count_radicals(mol: Mol) -> int:
    """Count the number of radicals in a molecule."""

    return sum(atom.GetNumRadicalElectrons() for atom in mol.GetAtoms())


def hydrate_radicals(mol: Mol) -> Mol:
    """Hydrate radicals in a molecule. Modifies molecule in place."""

    SanitizeMol(mol)
    for atom in mol.GetAtoms():
        num_atom_radicals = atom.GetNumRadicalElectrons()
        if num_atom_radicals:
            atom.SetNumExplicitHs(atom.GetNumExplicitHs() + num_atom_radicals)
            atom.SetNumRadicalElectrons(0)
    SanitizeMol(mol)
    return mol


def optimize_positions(mol: Mol, fixed: set[int]) -> tuple[Mol, float, float]:
    """Optimize the positions of the atoms in index set."""

    forcefield = UFFGetMoleculeForceField(mol, confId=0)
    e_start = forcefield.CalcEnergy()
    if mol.GetNumAtoms() == len(fixed):
        return mol, e_start, e_start
    for i in fixed:
        forcefield.AddFixedPoint(mol.GetAtomWithIdx(i).GetIdx())
    forcefield.Minimize()
    e_min = forcefield.CalcEnergy()
    return mol, e_start, e_min


def optimize_hydrogen_positions(mol: Mol) -> tuple[Mol, float, float]:
    """Optimize the positions of hydrogens."""
    heavy_atoms = {atom.GetIdx() for atom in mol.GetAtoms() if atom.GetAtomicNum() != 1}
    return optimize_positions(mol, heavy_atoms)


def add_hydrogens_with_uff_positions(mol: Mol) -> tuple[Mol, float, float]:
    SanitizeMol(mol)
    indices_before = {atom.GetIdx() for atom in mol.GetAtoms()}
    mol = hydrate_radicals(mol)
    mol = AddHs(mol, addCoords=True)
    return optimize_positions(mol, indices_before)


def _align_and_renumber(mol_true: Mol, mol_pred: Mol) -> Mol:
    """Align mol_pred to mol_true and renumbers atoms in mol_true to match mol_pred."""
    alignment = GetBestAlignmentTransform(mol_true, mol_pred, symmetrizeConjugatedTerminalGroups=False)
    # rotate and translate
    TransformConformer(mol_true.GetConformer(0), alignment[1])
    # renumber atoms
    map = [m[0] for m in sorted(alignment[2], key=lambda x: x[1])]  # reverse map
    mol_true = RenumberAtoms(mol_true, map)
    return mol_true


def _renumber_mol1_to_match_mol2(mol1: Mol, mol2: Mol) -> Mol:
    """Renumber atoms in mol1 to match mol2 (in terms of RMSD)."""
    rmsd, transform, atom_map = GetBestAlignmentTransform(mol1, mol2, reflect=False)
    is_identical_idx = [p[0] == p[1] for p in atom_map]
    n_atoms = mol1.GetNumAtoms()
    n_different = n_atoms - sum(is_identical_idx)
    if all(is_identical_idx):
        logger.info("All %d atoms are already in the same order", n_atoms)
    else:
        logger.info("Swapping %d out of %d indices", n_different, n_atoms)
        mol1 = RenumberAtoms(mol1, [m[1] for m in atom_map])
    return mol1


_periodic_table = GetPeriodicTable()


def _get_atomic_number(atomic_symbol: str):
    try:
        symbol = atomic_symbol[0].upper() + atomic_symbol[1:].lower()
        if symbol == "D":
            symbol = "H"
        return _periodic_table.GetAtomicNumber(symbol)
    except Exception:
        logger.error("Unknown atomic symbol: %s", atomic_symbol)
        return atomic_symbol


def _get_bond_length(mol: Mol, bond: Bond) -> float:
    b1, b2 = bond.GetBeginAtomIdx(), bond.GetEndAtomIdx()
    conf = mol.GetConformer()
    length = conf.GetAtomPosition(b1).Distance(conf.GetAtomPosition(b2))
    return float(length)


def _calc_angle(distances: np.ndarray, angle: tuple[int, int, int]) -> float:
    # calc alpha (opposite angle of side a in triangle with side lengths a, b, c)
    a = distances[angle[0], angle[2]]
    b = distances[angle[0], angle[1]]
    c = distances[angle[1], angle[2]]
    return float(np.degrees(np.arccos((b**2 + c**2 - a**2) / (2 * b * c))))



"""Module to check identity of docked and crystal ligand."""

from __future__ import annotations

import logging
from typing import Any

from rdkit.Chem.rdchem import Mol
from rdkit.rdBase import LogToPythonLogger

from ..tools.inchi import is_valid_inchi, split_inchi, standardize_and_get_inchi

LogToPythonLogger()
logger = logging.getLogger(__name__)


def check_identity(mol_pred: Mol, mol_true: Mol, inchi_options: str = "") -> dict[str, Any]:
    """Check two molecules are identical (docking relevant identity).

    Args:
        mol_pred: Predicted molecule (docked ligand).
        mol_true:Ground truth molecule (crystal ligand) with a conformer.
        inchi_options: String of options to pass to the InChI module. Defaults to "".

    Returns:
        PoseBusters results dictionary.
    """
    # generate inchis
    inchi_crystal = standardize_and_get_inchi(mol_true, options=inchi_options)
    inchi_docked = standardize_and_get_inchi(mol_pred, options=inchi_options)

    # check inchis are valid
    inchi_crystal_valid = is_valid_inchi(inchi_crystal)
    inchi_docked_valid = is_valid_inchi(inchi_docked)

    # compare inchis
    if inchi_crystal_valid and inchi_docked_valid:
        inchi_comparison = _compare_inchis(inchi_docked, inchi_crystal)
    else:
        inchi_comparison = {}

    results = {
        "inchi_crystal_valid": inchi_crystal_valid,
        "inchi_docked_valid": inchi_docked_valid,
        "inchi_crystal": inchi_crystal,
        "inchi_docked": inchi_docked,
        **inchi_comparison,
    }

    return {"results": results}


# isotopic layer /i (for isotopes) not present if no isotopic information in molecule
# fixed hydrogen layer /f (for tautomers) not part of standard inchi string
# reconnected layer /r (for metal ions) not part of standard inchi string
standard_layers = ["=", "/", "/c", "/h", "/q", "/p", "/t", "/b", "/m", "/s"]
layer_names = {
    "=": "inchi_version",
    "/": "formula",
    "/c": "connections",
    "/h": "hydrogens",
    "/q": "net_charge",
    "/p": "protons",
    "/b": "stereo_dbond",  # double bond (Z/E) stereochemistry
    "/t": "stereo_sp3",  # tetrahderal stereochemistry
    "/m": "stereo_sp3_inverted",
    "/s": "stereo_type",
    "/i": "isotopic",
}
stereo_all_layers = ["stereo_dbond", "stereo_sp3", "stereo_sp3_inverted", "stereo_type"]
stereo_tetrahedral_layers = ["stereo_sp3", "stereo_sp3_inverted", "stereo_type"]


def _compare_inchis(inchi_true: str, inchi_pred: str, layers: list[str] = standard_layers) -> dict[str, bool]:
    results = {}

    # fast return when overall InChI is the same
    if inchi_true == inchi_pred:
        results["inchi_overall"] = True
        for layer in layers:
            results[layer_names[layer]] = True
            results["stereo_tetrahedral"] = True
            results["stereo"] = True
        return results

    # otherwise comparison by layer
    results["inchi_overall"] = False
    layers_true = split_inchi(inchi_true)
    layers_pred = split_inchi(inchi_pred)
    assert "/" in layers_true, "Molecular formula layer missing from InChI string"
    for layer in layers:
        name = layer_names[layer]
        if (layer not in layers_true) or (layer not in layers_pred):
            results[name] = (layer not in layers_true) and (layer not in layers_pred)
        else:
            results[name] = layers_true[layer] == layers_pred[layer]

    # combine stereo layers (pass if not present)
    results["stereo_tetrahedral"] = all(results.get(name, True) for name in stereo_tetrahedral_layers)
    results["stereo"] = all(results.get(name, True) for name in stereo_all_layers)

    return results



from __future__ import annotations

from rdkit.Chem.rdmolfiles import MolFromSmiles

from posebusters.modules.identity import (
    check_identity,
    standardize_and_get_inchi,
)
from posebusters.tools.inchi import is_valid_inchi, split_inchi

mol_ethane = MolFromSmiles("CC")
mol_ethene = MolFromSmiles("C=C")
mol_ethyne = MolFromSmiles("C#C")

mol_benzene = MolFromSmiles("C1=CC=CC=C1")
mol_cyclohexane = MolFromSmiles("C1CCCCC1")

mol_ethylamine = MolFromSmiles("CCN")
mol_ethylaminium = MolFromSmiles("CC[NH3+]")
mol_ethanaminide = MolFromSmiles("CC[NH-]")

mol_sp3_anticlockwise = MolFromSmiles("N[C@H](C)C(=O)O")
mol_sp3_clockwise = MolFromSmiles("N[C@@H](C)C(=O)O")

mol_thiamine = MolFromSmiles("CC1=C(SC=[N+]1CC2=CN=C(N=C2N)C)CCO")


def test_check_identity_positive():
    out = check_identity(mol_benzene, mol_benzene)
    assert out["results"]["inchi_overall"] is True


def test_check_identity_negative():
    out = check_identity(mol_benzene, mol_cyclohexane)
    assert out["results"]["inchi_overall"] is False
    assert out["results"]["formula"] is False
    assert out["results"]["connections"] is True


def test_check_identity_tetrahedral():
    out = check_identity(mol_sp3_clockwise, mol_sp3_clockwise)
    assert out["results"]["inchi_overall"] is True
    assert out["results"]["stereo_type"] is True
    assert out["results"]["stereo_sp3"] is True
    assert out["results"]["stereo_sp3_inverted"] is True
    assert out["results"]["stereo_tetrahedral"] is True

    out = check_identity(mol_sp3_clockwise, mol_sp3_anticlockwise)
    assert out["results"]["inchi_overall"] is False
    assert out["results"]["stereo_type"] is True
    assert out["results"]["stereo_tetrahedral"] is False


def test_standardize_and_get_inchi():
    assert standardize_and_get_inchi(mol_ethane) == "InChI=1S/C2H6/c1-2/h1-2H3"

    # bond order matters
    assert standardize_and_get_inchi(mol_ethane) != standardize_and_get_inchi(mol_ethene)
    assert standardize_and_get_inchi(mol_benzene) != standardize_and_get_inchi(mol_cyclohexane)

    # protonation state to be normalized
    assert standardize_and_get_inchi(mol_ethylamine) == standardize_and_get_inchi(mol_ethylaminium)
    assert standardize_and_get_inchi(mol_ethylamine) == standardize_and_get_inchi(mol_ethanaminide)


def test_split_inchi():
    inchi_ethane = standardize_and_get_inchi(mol_ethane)
    split_inchi_ethane = split_inchi(inchi_ethane)
    assert split_inchi_ethane["/"] == "C2H6"
    assert split_inchi_ethane["/c"] == "1-2"
    assert split_inchi_ethane["/h"] == "1-2H3"


def test_is_valid_inchi():
    phenol_inchi = "InChI=1S/C6H6O/c7-6-4-2-1-3-5-6/h1-5,7H"
    assert is_valid_inchi(phenol_inchi) is True
    assert is_valid_inchi("1S/C6H6O/c7-6-4-2-1-3-5-6/h1-5,7H") is False
    assert is_valid_inchi("InChI=1S/c7-6-4-2-1-3-5-6/h1-5,7H") is False
    assert is_valid_inchi("InChI=1S/C6H6O/c7-6-4-1-3-5-6/h1-5,7H") is False
    assert is_valid_inchi("InChI=1S/C6H6O/c7-6-4-1-3-5-6") is False



from __future__ import annotations

import pytest

from posebusters.modules.rmsd import check_rmsd


def test_check_rmsd(mol_rq3_x00, mol_rq3_x01, mol_rq3_x10):
    out = check_rmsd(mol_rq3_x00, mol_rq3_x00)
    assert out["results"]["rmsd"] == pytest.approx(0.0, abs=1e-6)
    assert out["results"]["rmsd_within_threshold"] is True

    out = check_rmsd(mol_rq3_x00, mol_rq3_x01)
    assert out["results"]["rmsd"] == pytest.approx(1.0, abs=1e-6)

    out = check_rmsd(mol_rq3_x00, mol_rq3_x10)
    assert out["results"]["rmsd"] == pytest.approx(10.0, abs=1e-6)


def test_check_rmsd_1mmv_3ar(mol_true_1mmv_3ar, mol_pred_1mmv_3ar):
    out = check_rmsd(mol_true_1mmv_3ar, mol_pred_1mmv_3ar)
    assert out["results"]["rmsd"] < 10.0
    assert out["results"]["rmsd_within_threshold"] is False


def test_check_rmsd_1of6_dty(mol_true_1of6_dty, mol_pred_1of6_dty):
    out = check_rmsd(mol_true_1of6_dty, mol_pred_1of6_dty)
    assert out["results"]["rmsd"] < 2.0
    assert out["results"]["rmsd_within_threshold"] is True


def test_check_rmsd_1q1g_mti(mol_true_1q1g_mti, mol_pred_1q1g_mti):
    out = check_rmsd(mol_true_1q1g_mti, mol_pred_1q1g_mti)
    assert out["results"]["rmsd"] < 50.0
    assert out["results"]["rmsd_within_threshold"] is False


def test_check_rmsd_6yr2_t1c(mol_true_6yr2_t1c, mol_pred_6yr2_t1c):
    out = check_rmsd(mol_true_6yr2_t1c, mol_pred_6yr2_t1c)
    assert out["results"]["rmsd"] < 50.0
    assert out["results"]["rmsd_within_threshold"] is False


def test_check_rmsd_tautomeric(mol_pred_1g9v, mol_true_1g9v):
    out = check_rmsd(mol_pred_1g9v, mol_true_1g9v)
    assert out["results"]["rmsd"] == pytest.approx(6.97287, abs=1e-4)


def test_check_rmsd_multiple_ground_truth(mol_one_true_1w1p, mol_true_1w1p):
    out = check_rmsd(mol_one_true_1w1p, mol_true_1w1p)
    assert out["results"]["rmsd"] == pytest.approx(0.0, abs=1e-6)
    assert out["results"]["rmsd_within_threshold"] is True


def test_check_rmsd_bond_orders(mol_3wrb_gde_true, mol_3wrb_gde_pred):
    out = check_rmsd(mol_3wrb_gde_true, mol_3wrb_gde_true)
    assert out["results"]["rmsd"] == pytest.approx(0.0, abs=1e-6)

    out = check_rmsd(mol_3wrb_gde_pred, mol_3wrb_gde_pred)
    assert out["results"]["rmsd"] == pytest.approx(0.0, abs=1e-6)

    out = check_rmsd(mol_3wrb_gde_pred, mol_3wrb_gde_true)
    assert out["results"]["rmsd"] == pytest.approx(1.091668, abs=1e-4)

    out = check_rmsd(mol_3wrb_gde_true, mol_3wrb_gde_pred)
    assert out["results"]["rmsd"] == pytest.approx(1.091668, abs=1e-4)



"""Logging utilities."""

from __future__ import annotations

import logging
import os
import sys
from typing import Any

import rdkit
from rdkit import rdBase

# redirect logs to Python logger
rdBase.LogToPythonLogger()


# https://github.com/rdkit/rdkit/discussions/5435
class CaptureLogger(logging.Handler):
    """Helper class that captures Python logger output."""

    def __init__(self, level: str | None = None) -> None:
        """Initialize logger."""
        super().__init__(level=logging.NOTSET if level is None else level)
        self.logs: dict[str, str] = {}
        self.devnull = open(os.devnull, "w")
        rdkit.log_handler.setStream(self.devnull)
        rdkit.logger.addHandler(self)

    def __enter__(self) -> dict[str, str]:
        """Enter context manager."""
        return self.logs

    def __exit__(self, exc_type: type[BaseException] | None, exc_val: BaseException | None, exc_tb: Any) -> None:
        """Exit context manager."""
        self.release()

    def handle(self, record: logging.LogRecord) -> bool:
        """Handle log record."""
        key = record.levelname
        val = self.format(record)
        self.logs[key] = self.logs.get(key, "") + val
        return False

    def release(self) -> None:
        """Release logger."""
        rdkit.log_handler.setStream(sys.stderr)
        rdkit.logger.removeHandler(self)
        self.devnull.close()

    def emit(self, record: logging.LogRecord) -> None:
        """Emit a log record."""



import math
from pathlib import Path

from rdkit.Chem.rdmolfiles import MolFromMolFile, MolFromPDBFile, MolFromSmiles
from rdkit.Chem.rdmolops import AddHs, SanitizeMol

from posebusters import PoseBusters

mols_table = "tests/conftest/sample_bust_docks_table.csv"

mol_conf_2 = "tests/conftest/mol_1a30_clash_2.sdf"
mol_conf_3 = "tests/conftest/mol_1a30_clash_3.sdf"

mol_pred_1ia1 = "tests/conftest/1IA1_TQ3/1IA1_TQ3_ligand.sdf"
mol_true_1ia1 = "tests/conftest/1IA1_TQ3/1IA1_TQ3_ligands.sdf"
mol_cond_1ia1 = "tests/conftest/1IA1_TQ3/1IA1_TQ3_no_ligand_no_solvent.pdb"

mol_pred_1w1p = "tests/conftest/1W1P_GIO/1W1P_GIO_ligand.sdf"
mol_true_1w1p = "tests/conftest/1W1P_GIO/1W1P_GIO_ligands.sdf"
mol_cond_1w1p = "tests/conftest/1W1P_GIO/1W1P_GIO_protein.pdb"

mol_pred_7w2p = "tests/conftest/7W2P_8AI/7W2P_8AI_ligand.sdf"
mol_true_7w2p = "tests/conftest/7W2P_8AI/7W2P_8AI_ligands.sdf"
mol_cond_7w2p = "tests/conftest/7W2P_8AI/7W2P_8AI_no_ligand_no_solvent.pdb"

mol_single_h = "tests/conftest/mol_single_hydrogen.sdf"

mol_smaller = "tests/conftest/2HA2_SCK_2HA3_CHT/2HA2_SCK_2HA3_CHT_smaller_ligand.sdf"
mol_larger = "tests/conftest/2HA2_SCK_2HA3_CHT/2HA2_SCK_2HA3_CHT_larger_ligand.sdf"
mol_cond_smaller = "tests/conftest/2HA2_SCK_2HA3_CHT/2HA2_SCK_2HA3_CHT_smaller_receptor.pdb"

mol_true_5ze6 = "tests/conftest/5ze6/5ze6_true.mol2"
mol_pred_5ze6 = "tests/conftest/5ze6/5ze6_pred.sdf"
mol_cond_5ze6 = "tests/conftest/5ze6/5ze6_cond.pdb"

mol_true_sanity = "tests/conftest/sanity_error/true.sdf"
mol_pred_sanity = "tests/conftest/sanity_error/pred.sdf"
mol_cond_sanity = "tests/conftest/sanity_error/protein.pdb"

mol_true_sanity_2 = "tests/conftest/sanity_error_2/true.sdf"
mol_pred_sanity_2 = "tests/conftest/sanity_error_2/pred.sdf"
mol_cond_sanity_2 = "tests/conftest/sanity_error_2/protein.pdb"


def test_bust_redocks_1ia1() -> None:
    posebusters = PoseBusters("redock")
    df = posebusters.bust([mol_pred_1ia1], mol_true_1ia1, mol_cond_1ia1)
    assert df.all(axis=1).values[0]


def test_bust_redocks_1w1p() -> None:
    posebusters = PoseBusters("redock")
    df = posebusters.bust([mol_pred_1w1p], mol_true_1w1p, mol_cond_1w1p)
    assert df.all(axis=1).values[0]


def test_bust_redocks_5ze6() -> None:
    # check that mol2 files as true molecule can be loaded

    posebusters = PoseBusters("redock")
    df = posebusters.bust([mol_pred_5ze6], mol_true_5ze6, mol_cond_5ze6)
    assert df["mol_true_loaded"].all()


def test_bust_docks() -> None:
    posebusters = PoseBusters("dock")
    df = posebusters.bust([mol_pred_1ia1], mol_cond=mol_cond_1ia1)
    assert df.all(axis=1).values[0]


def test_bust_mols() -> None:
    posebusters = PoseBusters("mol")

    # pass one not in list
    df = posebusters.bust(mol_pred_1ia1)
    assert df.all(axis=1).values[0]

    # pass list
    df = posebusters.bust([mol_pred_1ia1])
    assert df.all(axis=1).values[0]


def test_bust_mol_rdkit() -> None:
    posebusters = PoseBusters(config="mol")
    mol = MolFromSmiles("C")

    df = posebusters.bust(mol)
    assert df.all(axis=1).values[0]

    df = posebusters.bust([mol])
    assert df.all(axis=1).values[0]


def test_bust_mols_hydrogen(threshold=8) -> None:
    posebusters = PoseBusters("mol")
    df = posebusters.bust([mol_single_h])
    assert df.sum(axis=1).values[0] >= threshold  # energy ratio test fails


def test_bust_mols_consistency(atol=1e-6) -> None:
    # check that running the same molecule twice gives the same result

    posebusters = PoseBusters("mol")
    result_2 = posebusters.bust([mol_conf_2])

    posebusters = PoseBusters("mol")
    result_3 = posebusters.bust([mol_conf_2])

    # check numbers are approximately equal
    for v1, v2 in zip(result_2.values, result_3.values):
        if v1[2] == v2[2] or math.isnan(v1[2]):
            continue
        assert abs(v1[2] - v2[2]) < atol, f"{v1[0], v1[1]}: {v1[2]} != {v2[2]}"


def test_bust_gen() -> None:
    posebusters = PoseBusters("gen")
    result = posebusters.bust(mol_pred=mol_larger, mol_true=mol_smaller, mol_cond=mol_cond_smaller, full_report=True)
    assert result["sucos"].iloc[0] > 0.2


def test_bust_loaded_mols() -> None:
    posebuster = PoseBusters("redock")

    mol_pred = MolFromMolFile(mol_pred_1ia1)
    mol_true = MolFromMolFile(mol_true_1ia1)
    mol_cond = MolFromPDBFile(mol_cond_1ia1)

    df = posebuster.bust([mol_pred], mol_true, mol_cond)
    assert df["mol_pred_loaded"].all()
    assert df["mol_true_loaded"].all()
    assert df["mol_cond_loaded"].all()


def test_check_energy_ratio_14gs_0(mol_pred_14gs_gen0):
    posebuster = PoseBusters("mol")
    df = posebuster.bust(mol_pred_14gs_gen0)

    assert df["mol_pred_loaded"].all()
    # assert that there is not a NA value given everyything else is correct
    assert not (df.eq(False) | df.isna()).all(axis=1).any()


def test_check_energy_ratio_1afs_87(mol_pred_1afs_gen87):
    posebuster = PoseBusters("mol")
    df = posebuster.bust(mol_pred_1afs_gen87)

    assert df["mol_pred_loaded"].all()
    # assert that there is not a NA value given everyything else is correct
    assert not (df.eq(False) | df.isna()).all(axis=1).any()


def test_check_energy_ratio_1afs_94(mol_pred_1afs_gen94):
    posebuster = PoseBusters("mol")
    df = posebuster.bust(mol_pred_1afs_gen94)

    assert df["mol_pred_loaded"].all()
    # assert that there is not a NA value given everyything else is correct
    assert not (df.eq(False) | df.isna()).all(axis=1).any()


def test_check_energy_ratio_1jn2_3(mol_pred_1jn2_gen3):
    posebuster = PoseBusters("mol")
    df = posebuster.bust(mol_pred_1jn2_gen3)

    assert df["mol_pred_loaded"].all()
    # assert that there is not a NA value given everyything else is correct
    assert not (df.eq(False) | df.isna()).all(axis=1).any()


def test_check_energy_ratio_1jn2_62(mol_pred_1jn2_gen62):
    posebuster = PoseBusters("mol")
    df = posebuster.bust(mol_pred_1jn2_gen62)

    assert df["mol_pred_loaded"].all()
    # assert that there is not a NA value given everyything else is correct
    assert not (df.eq(False) | df.isna()).all(axis=1).any()


def test_check_sanity():
    posebusters = PoseBusters("redock")
    df = posebusters.bust([mol_pred_sanity], mol_true_sanity, mol_cond_sanity, full_report=True)
    assert df["mol_true_loaded"].all()
    assert (df["rmsd"] < 3).all()


def test_check_sanity_2():
    posebusters = PoseBusters("redock")
    df = posebusters.bust([mol_pred_sanity_2], mol_true_sanity_2, mol_cond_sanity_2, full_report=True)
    assert df["mol_true_loaded"].all()
    assert (df["rmsd"] < 3).all()


def test_parallel_with_path():
    posebusters = PoseBusters("mol", max_workers=3)
    df = posebusters.bust(mol_pred=Path("tests/conftest/mol_1jn2_62_out.sdf"))
    assert df["mol_pred_loaded"].all()


def test_parallel_with_mol(mol_pred_1jn2_gen62):
    posebusters = PoseBusters("mol", max_workers=3)
    df = posebusters.bust(mol_pred=mol_pred_1jn2_gen62)
    assert df["mol_pred_loaded"].all()


def test_issue_67(mol_issue_67):
    # test the molecule that caused issue #67

    # mol without sanitization
    buster = PoseBusters(config="mol", max_workers=0)
    pb_result = buster.bust([mol_issue_67], None, None)
    assert pb_result["bond_lengths"].iloc[0]

    # mol with sanitization
    SanitizeMol(mol_issue_67, catchErrors=True)
    buster = PoseBusters(config="mol", max_workers=0)
    pb_result = buster.bust([mol_issue_67], None, None)
    assert pb_result["bond_lengths"].iloc[0]

    # mol with added hydrogens
    mol_issue_67 = AddHs(mol_issue_67)
    buster = PoseBusters(config="mol", max_workers=0)
    pb_result = buster.bust([mol_issue_67], None, None)
    assert pb_result["bond_lengths"].iloc[0]



from __future__ import annotations

import pytest
from rdkit.Chem.rdchem import Conformer, Mol
from rdkit.Chem.rdDistGeom import EmbedMolecule, srETKDGv3
from rdkit.Chem.rdmolfiles import MolFromMol2File, MolFromMolFile, MolFromPDBFile, MolFromSmiles, SDMolSupplier
from rdkit.Chem.rdmolops import AddHs


@pytest.fixture
def mol_cgb():
    return MolFromMolFile("tests/conftest/mol_CGB.sdf")


@pytest.fixture
def mol_pm2():
    return MolFromMolFile("tests/conftest/mol_PM2.sdf")


@pytest.fixture
def mol_rq3_x00():
    return MolFromMolFile("tests/conftest/mol_RQ3_x00.sdf")


@pytest.fixture
def mol_rq3_x01():
    return MolFromMolFile("tests/conftest/mol_RQ3_x01.sdf")


@pytest.fixture
def mol_rq3_x10():
    return MolFromMolFile("tests/conftest/mol_RQ3_x10.sdf")


@pytest.fixture
def ligand_2bm2():
    return MolFromMolFile("tests/conftest/mol_2bm2_ligand.sdf")


@pytest.fixture
def protein_2bm2():
    return MolFromPDBFile(
        "tests/conftest/mol_2bm2_protein_one_ligand_removed.pdb", sanitize=False, removeHs=False, proximityBonding=False
    )


@pytest.fixture
def mol_1a30_ligand():
    return MolFromMol2File("tests/conftest/mol_1a30_ligand.mol2", sanitize=False)


@pytest.fixture
def mol_1a30_clash_2():
    return MolFromMolFile("tests/conftest/mol_1a30_clash_2.sdf", sanitize=False)


@pytest.fixture
def mol_1a30_clash_3():
    return MolFromMolFile("tests/conftest/mol_1a30_clash_3.sdf", sanitize=False)


@pytest.fixture
def mol_pred_14gs_gen0():
    return MolFromMolFile("tests/conftest/mol_14gs_0_out.sdf", sanitize=True)


@pytest.fixture
def mol_pred_1afs_gen87():
    return MolFromMolFile("tests/conftest/mol_1afs_87_out.sdf", sanitize=True)


@pytest.fixture
def mol_pred_1afs_gen94():
    return MolFromMolFile("tests/conftest/mol_1afs_94_out.sdf", sanitize=True)


@pytest.fixture
def mol_pred_1jn2_gen3():
    return MolFromMolFile("tests/conftest/mol_1jn2_3_out.sdf", sanitize=True)


@pytest.fixture
def mol_pred_1jn2_gen62():
    return MolFromMolFile("tests/conftest/mol_1jn2_62_out.sdf", sanitize=True)


@pytest.fixture
def mol_calcium():
    return MolFromMolFile("tests/conftest/mol_calcium.sdf", sanitize=True)


@pytest.fixture
def mol_cholesterol():
    return MolFromMolFile("tests/conftest/mol_cholesterol.sdf", sanitize=True)


@pytest.fixture
def mol_true_1g9v():
    path = "tests/conftest/1G9V_RQ3/1G9V_RQ3_ligands.sdf"
    with SDMolSupplier(str(path), sanitize=True, removeHs=True, strictParsing=True) as supplier:
        mol = next(supplier)
        while mol is None and supplier.atEnd() is False:
            mol = next(supplier)
        for mol_next in supplier:
            if mol_next is not None:
                mol.AddConformer(mol_next.GetConformer(), assignId=True)
        return mol


@pytest.fixture
def mol_one_true_1g9v():
    return MolFromMolFile("tests/conftest/1G9V_RQ3/1G9V_RQ3_ligand.sdf", sanitize=True)


@pytest.fixture
def mol_pred_1g9v():
    return MolFromMolFile("tests/conftest/1G9V_RQ3/1G9V_RQ3_gold_redock.sdf", sanitize=True)


@pytest.fixture
def mol_true_1w1p():
    path = "tests/conftest/1W1P_GIO/1W1P_GIO_ligands.sdf"
    with SDMolSupplier(str(path), sanitize=True, removeHs=True, strictParsing=True) as supplier:
        mol = next(supplier)
        while mol is None and supplier.atEnd() is False:
            mol = next(supplier)
        for mol_next in supplier:
            if mol_next is not None:
                mol.AddConformer(mol_next.GetConformer(), assignId=True)
        return mol


@pytest.fixture
def mol_one_true_1w1p():
    return MolFromMolFile("tests/conftest/1W1P_GIO/1W1P_GIO_ligand.sdf", sanitize=True)


@pytest.fixture
def mol_true_1mmv_3ar():
    return MolFromMolFile("tests/conftest/1MMV_3AR/1MMV_3AR_crystal.sdf", sanitize=True)


@pytest.fixture
def mol_pred_1mmv_3ar():
    return MolFromMolFile("tests/conftest/1MMV_3AR/1MMV_3AR_tankbind.sdf", sanitize=True)


@pytest.fixture
def mol_true_1of6_dty():
    return MolFromMolFile("tests/conftest/1OF6_DTY/1OF6_DTY_crystal.sdf", sanitize=True)


@pytest.fixture
def mol_pred_1of6_dty():
    return MolFromMolFile("tests/conftest/1OF6_DTY/1OF6_DTY_vina.sdf", sanitize=True)


@pytest.fixture
def mol_true_1q1g_mti():
    return MolFromMolFile("tests/conftest/1Q1G_MTI/1Q1G_MTI_crystal.sdf", sanitize=True)


@pytest.fixture
def mol_pred_1q1g_mti():
    return MolFromMolFile("tests/conftest/1Q1G_MTI/1Q1G_MTI_tankbind.sdf", sanitize=True)


@pytest.fixture
def mol_true_6yr2_t1c():
    return MolFromMolFile("tests/conftest/6YR2_T1C/6YR2_T1C_crystal.sdf", sanitize=True)


@pytest.fixture
def mol_pred_6yr2_t1c():
    return MolFromMolFile("tests/conftest/6YR2_T1C/6YR2_T1C_tankbind.sdf", sanitize=True)


@pytest.fixture
def mol_lig_7ecr_sin():
    return MolFromMolFile("tests/conftest/7ECR_SIN/7ECR_SIN_ligand.sdf", sanitize=True)


@pytest.fixture
def mol_cond_7ecr_sin():
    return MolFromPDBFile("tests/conftest/7ECR_SIN/7ECR_SIN_protein.pdb", sanitize=False, proximityBonding=False)


@pytest.fixture
def mol_lig_7cnq_g8x():
    return MolFromMolFile("tests/conftest/7CNQ_G8X/7CNQ_G8X_ligand.sdf", sanitize=True)


@pytest.fixture
def mol_cond_7cnq_g8x():
    return MolFromPDBFile("tests/conftest/7CNQ_G8X/7CNQ_G8X_protein.pdb", sanitize=False, proximityBonding=False)


@pytest.fixture
def mol_lig_7ztl_bcn():
    return MolFromMolFile("tests/conftest/7ZTL_BCN/7ZTL_BCN_ligand.sdf", sanitize=True)


@pytest.fixture
def mol_cond_7ztl_bcn():
    return MolFromPDBFile("tests/conftest/7ZTL_BCN/7ZTL_BCN_protein.pdb", sanitize=False, proximityBonding=False)


@pytest.fixture
def mol_disconnnected_atoms():
    return MolFromMolFile("tests/conftest/mol_disconnected_atoms.sdf", sanitize=True)


@pytest.fixture
def mol_small_7brv_f5r_7wb6_f5r():
    return MolFromMolFile("tests/conftest/7BRV_F5R_7WB6_F5R/7BRV_F5R_7WB6_F5R_smaller_ligand.sdf", removeHs=True)


@pytest.fixture
def mol_large_7brv_f5r_7wb6_f5r():
    return MolFromMolFile("tests/conftest/7BRV_F5R_7WB6_F5R/7BRV_F5R_7WB6_F5R_larger_ligand.sdf", removeHs=True)


@pytest.fixture
def mol_065():
    return MolFromMolFile("tests/conftest/mol_065_ideal.sdf")


@pytest.fixture
def mol_065_left():
    return MolFromMolFile("tests/conftest/mol_065_left.sdf")


@pytest.fixture
def mol_065_right():
    return MolFromMolFile("tests/conftest/mol_065_right.sdf")


@pytest.fixture
def mol_TMO():
    return MolFromMolFile("tests/conftest/mol_TMO.sdf")


@pytest.fixture
def mol_2YU():
    return MolFromMolFile("tests/conftest/mol_2YU.mol")


@pytest.fixture
def mol_HQT():
    return MolFromMolFile("tests/conftest/mol_HQT.mol")


@pytest.fixture
def mol_3wrb_gde_true():
    return MolFromPDBFile("tests/conftest/mol_3WRB_1_GDE_0_ligand_true.pdb")


@pytest.fixture
def mol_3wrb_gde_pred():
    return MolFromMolFile("tests/conftest/mol_3WRB_1_GDE_0_ligand_pred.sdf")


def embed_mol(smi: str) -> Mol:
    hmol = AddHs(MolFromSmiles(smi))
    ps = srETKDGv3()
    ps.randomSeed = 42
    _ = EmbedMolecule(hmol, params=ps)
    return hmol


@pytest.fixture()
def mols_flat_etkdgv3():
    # non-aromatic flat
    smis_flat = ["C1=CC2=CC=CC2=C1", "C1=C[CH]C=C1", "N1C=CNC=C1", "C1C=CN=CN1"]
    # aromatic flat
    smis_flat.extend(["c1ccccc1", "c1cnc[nH]1"])
    return [embed_mol(smi) for smi in smis_flat]


@pytest.fixture()
def mols_nonflat_etkdgv3():
    # just nonflat
    smis_flat = ["C1CC(C)NC1", "N1C=CCC=C1", "C1C=CCCO1"]
    return [embed_mol(smi) for smi in smis_flat]


@pytest.fixture()
def mol_awj_ideal():
    return MolFromMolFile("tests/conftest/mol_AWJ_ideal.sdf")


@pytest.fixture()
def mol_o2u_ideal():
    return MolFromMolFile("tests/conftest/mol_O2U_ideal.sdf")


@pytest.fixture()
def mol_pip_ideal():
    return MolFromMolFile("tests/conftest/mol_PIP_ideal.sdf")


@pytest.fixture()
def mol_pip_wrong():
    return MolFromMolFile("tests/conftest/mol_PIP_wrong.sdf")


@pytest.fixture()
def mol_issue_67():
    return MolFromMolFile("tests/conftest/mol_issue_67.sdf")


@pytest.fixture
def mol_fe2plus():
    mol = MolFromSmiles("[Fe+2]")
    conformer = Conformer(1)
    conformer.SetAtomPosition(0, [0.0, 0.0, 0.0])
    mol.AddConformer(conformer)
    return mol



"""Functions for formatting the output of the PoseBusters tools."""

import pandas as pd


def _value_map(x: bool) -> str:
    if isinstance(x, bool):
        return ".   " if x else "Fail"
    return x


def create_long_output(df: pd.DataFrame) -> str:
    """Create a detailed output string for the long PoseBusters reporting format.

    Args:
        df: Report dataframe.

    Returns:
        String with detailed report.
    """
    df = df.T
    df[df.columns[-1]] = df[df.columns[-1]].map(_value_map, na_action="ignore")
    cols = df.columns
    output = ""
    segments = []
    for col in cols:
        segments.append(
            "Long summary for " + " ".join(map(str, col)) + "\n" + df[[col]].to_string(index=True, header=False)
        )
    output += "\n\n".join(segments)
    return output + "\n"


def create_short_output(df: pd.DataFrame) -> str:
    """Create a one line output string for the short PoseBusters reporting format.

    Args:
        df: Report dataframe.

    Returns:
        String with one line report.
    """
    results = df.copy()
    results.columns = results.columns.to_flat_index()
    columns = results.columns
    passes = results[columns].sum(axis=1)
    results["Passed tests"] = passes.apply(lambda x: f"passes ({x} / {len(columns)})")
    results["Pass"] = results[columns].all(axis=1)
    results.index.name = None
    results = results[["Passed tests"]]
    return str(results.to_string(index=True, header=False)) + "\n"



"""PoseBusters: Plausibility checks for generated molecule poses."""

from posebusters.modules.distance_geometry import check_geometry
from posebusters.modules.energy_ratio import check_energy_ratio
from posebusters.modules.flatness import check_flatness
from posebusters.modules.identity import check_identity
from posebusters.modules.intermolecular_distance import check_intermolecular_distance
from posebusters.modules.loading import check_loading
from posebusters.modules.rmsd import check_rmsd
from posebusters.modules.sanity import (
    check_all_atoms_connected,
    check_chemistry,
    check_chemistry_using_inchi,
    check_chemistry_using_rdkit,
    check_radicals,
)
from posebusters.modules.volume_overlap import check_volume_overlap
from posebusters.posebusters import PoseBusters

__all__ = [
    "PoseBusters",
    "check_all_atoms_connected",
    "check_chemistry_using_inchi",
    "check_chemistry_using_rdkit",
    "check_chemistry",
    "check_energy_ratio",
    "check_flatness",
    "check_geometry",
    "check_identity",
    "check_intermolecular_distance",
    "check_loading",
    "check_radicals",
    "check_rmsd",
    "check_volume_overlap",
]

__version__ = "0.6.5"



"""Module to calculate SuCOS score."""

from __future__ import annotations

import os

import numpy as np
from rdkit import RDConfig
from rdkit.Chem.FeatMaps import FeatMaps
from rdkit.Chem.rdchem import Mol
from rdkit.Chem.rdMolChemicalFeatures import BuildFeatureFactory
from rdkit.Chem.rdmolops import AddHs, RemoveHs, SanitizeMol
from rdkit.Chem.rdShapeHelpers import ShapeProtrudeDist

FACTORY = BuildFeatureFactory(os.path.join(RDConfig.RDDataDir, "BaseFeatures.fdef"))
PARAMETERS = {k: FeatMaps.FeatMapParams() for k in FACTORY.GetFeatureFamilies()}
KEEP = (
    "Donor",
    "Acceptor",
    "NegIonizable",
    "PosIonizable",
    "ZnBinder",
    "Aromatic",
    "Hydrophobe",
    "LumpedHydrophobe",
)


def get_feature_map_score(
    mol_small: Mol,
    mol_large: Mol,
    conf_id_small: int = -1,
    conf_id_large: int = -1,
) -> float:
    """Calculate the feature map score between two molecules.

    References:
        Putta et al, 2005: https://pubs.acs.org/doi/abs/10.1021/jm049066l
        Landrum et al, 2007: https://link.springer.com/article/10.1007/s10822-006-9085-8
        Landrum Greg, 2017: https://rdkit.blogspot.com/2017/11/using-feature-maps.html
        Landrum Greg, 2023: https://greglandrum.github.io/rdkit-blog/posts/2023-02-24-using-feature-maps.html
    """

    # list features
    features_small = [f for f in FACTORY.GetFeaturesForMol(mol_small, confId=conf_id_small) if f.GetFamily() in KEEP]
    features_large = [f for f in FACTORY.GetFeaturesForMol(mol_large, confId=conf_id_large) if f.GetFamily() in KEEP]

    # create feature map based on small molecule
    feature_map = FeatMaps.FeatMap(feats=features_small, weights=[1] * len(features_small), params=PARAMETERS)
    feature_map.scoreMode = FeatMaps.FeatMapScoreMode.Best  # type: ignore[misc]

    # score features of large molecule present in small molecule
    feature_score = feature_map.ScoreFeats(features_large)

    # normalize score
    normalization_constant = min(feature_map.GetNumFeatures(), len(features_large))
    if normalization_constant > 0:
        return float(feature_score / normalization_constant)

    return np.nan


def get_sucos_score(
    mol_reference: Mol,
    mol_probe: Mol,
    conf_id_reference: int = -1,
    conf_id_probe: int = -1,
    heavy_only: bool | None = True,
) -> float:
    """Calculate the SuCOS score between a reference ligand and a list of probe ligands.

    Args:
        mol_reference: The smaller or reference molecule.
        mol_probe: The larger or query molecule.

    Returns:
        SuCOS score.

    References:
        Leung et al, 2019: https://chemrxiv.org/engage/chemrxiv/article-details/60c741a99abda23230f8bed5
        Moesser, Marc: https://github.com/MarcMoesser/SuCOS/blob/master/calc_SuCOS_normalized.py
    """

    # explicit or implicit hydrogens should be same for both molecules
    mol_reference = handle_hydrogens(mol_reference, heavy_only=heavy_only)
    mol_probe = handle_hydrogens(mol_probe, heavy_only=heavy_only)

    feature_map_score = get_feature_map_score(
        mol_small=mol_reference,
        mol_large=mol_probe,
        conf_id_small=conf_id_reference,
        conf_id_large=conf_id_probe,
    )
    protrusion_distance = ShapeProtrudeDist(
        mol1=mol_reference,
        mol2=mol_probe,
        confId1=conf_id_reference,
        confId2=conf_id_probe,
        allowReordering=False,
        vdwScale=0.8,
        ignoreHs=True,
    )
    shape_overlap = max(1 - protrusion_distance, 0)

    # if no features, base on shape alone
    if not np.isnan(feature_map_score):
        sucos_score = 0.5 * feature_map_score + 0.5 * shape_overlap
    else:
        sucos_score = shape_overlap

    sucos_score = min(max(0.0, sucos_score), 1.0)  # clipping
    return sucos_score


def check_sucos(mol_pred: Mol, mol_true: Mol, sucos_threshold: float = 0.4) -> dict[str, dict[str, bool | float]]:
    """Calculate SuCOS and related metrics between predicted molecule and closest ground truth molecule.

    Args:
        mol_pred: Predicted molecule (docked ligand) with exactly one conformer.
        mol_true: Ground truth molecule (crystal ligand) with at least one conformer. If multiple conformers are
            present, the highest SuCOS will be reported.
        sucos_threshold: Threshold in angstrom for reporting whether SuCOS is within threshold. Defaults to 0.4.
        heavy_only: Whether to only consider heavy atoms for SuCOS calculation. Defaults to True.

    Returns:
        PoseBusters results dictionary.
    """
    assert isinstance(mol_true, Mol), "Ground truth molecule is missing."
    assert isinstance(mol_pred, Mol), "Predicted molecule is missing."
    num_conf = mol_true.GetNumConformers()
    assert num_conf > 0, "Ground truth molecule needs at least one conformer."
    assert mol_pred.GetNumConformers() == 1, "Predicted molecule should only have one conformer."

    try:
        SanitizeMol(mol_pred)
        SanitizeMol(mol_true)
    except Exception:
        return {"results": {"sucos": np.nan, "sucos_within_threshold": np.nan}}

    # iterate over all true molecules to find best sucos match
    sucos_scores = [get_sucos_score(mol_true, mol_pred, conf_id_reference=i) for i in range(num_conf)]
    best_sucos = max(sucos_scores)
    sucos_within_threshold = best_sucos >= sucos_threshold

    results = {"sucos": best_sucos, "sucos_within_threshold": sucos_within_threshold}
    return {"results": results}


def handle_hydrogens(mol: Mol, heavy_only: bool | None = True) -> Mol:
    """Remove, add or do not modify hydrogens in a molecule."""
    if heavy_only is None:
        return mol
    if heavy_only:
        return RemoveHs(mol)
    return AddHs(mol, addCoords=True)



from __future__ import annotations

from posebusters.modules.intermolecular_distance import check_intermolecular_distance


def test_check_intermolecular_distance(protein_2bm2, ligand_2bm2):
    out = check_intermolecular_distance(ligand_2bm2, protein_2bm2)
    out["results"]["no_clashes"] is True


def test_check_intermolecular_distance_7ecr_sin(mol_lig_7ecr_sin, mol_cond_7ecr_sin):
    out = check_intermolecular_distance(
        mol_lig_7ecr_sin,
        mol_cond_7ecr_sin,
        radius_type="vdw",
        radius_scale=1.0,
        clash_cutoff=0.75,
        ignore_types={"organic_cofactors", "inorganic_cofactors"},
    )
    out["results"]["no_clashes"] is True



"""Module to check bond lengths, bond angles, and internal clash of ligand conformations."""

from __future__ import annotations

from collections.abc import Iterable
from copy import deepcopy
from logging import getLogger
from typing import Any

import numpy as np
import pandas as pd
from rdkit.Chem import MolFromSmarts
from rdkit.Chem.rdchem import Mol
from rdkit.Chem.rdDistGeom import GetMoleculeBoundsMatrix
from rdkit.Chem.rdmolops import SanitizeMol

from ..tools.stats import bape, bpe, pe

logger = getLogger(__name__)

col_lb = "lower_bound"
col_ub = "upper_bound"
col_pe = "percent_error"
col_bpe = "bound_percent_error"
col_bape = "bound_absolute_percent_error"

bound_matrix_params = {
    "set15bounds": True,
    "scaleVDW": True,
    "doTriangleSmoothing": True,
    "useMacrocycle14config": False,
}

col_n_bonds = "number_bonds"
col_shortest_bond = "shortest_bond_relative_length"
col_longest_bond = "longest_bond_relative_length"
col_n_short_bonds = "number_short_outlier_bonds"
col_n_long_bonds = "number_long_outlier_bonds"
col_n_good_bonds = "number_valid_bonds"
col_bonds_result = "bond_lengths_within_bounds"
col_n_angles = "number_angles"
col_extremest_angle = "most_extreme_relative_angle"
col_n_bad_angles = "number_outlier_angles"
col_n_good_angles = "number_valid_angles"
col_angles_result = "bond_angles_within_bounds"
col_n_noncov = "number_noncov_pairs"
col_closest_noncov = "shortest_noncovalent_relative_distance"
col_n_clashes = "number_clashes"
col_n_good_noncov = "number_valid_noncov_pairs"
col_clash_result = "no_internal_clash"

_empty_results = {
    col_n_bonds: np.nan,
    col_shortest_bond: np.nan,
    col_longest_bond: np.nan,
    col_n_short_bonds: np.nan,
    col_n_long_bonds: np.nan,
    col_bonds_result: np.nan,
    col_n_angles: np.nan,
    col_extremest_angle: np.nan,
    col_n_bad_angles: np.nan,
    col_angles_result: np.nan,
    col_n_noncov: np.nan,
    col_closest_noncov: np.nan,
    col_n_clashes: np.nan,
    col_clash_result: np.nan,
}


def symmetrize_conjugated_terminal_bonds(df: pd.DataFrame, mol: Mol) -> pd.DataFrame:
    """
    Symmetrize the lower and upper bounds of the conjugated terminal bonds so that
    the new lower and upper bounds are the minimum and maximum of the original
    lower and upper bounds for each pair of atom elements.

    Args:
        df: Dataframe with the bond geometry information and bounds.
        mol: RDKit molecule object (conformer id doesn't matter I think)

    Returns:
        Dataframe with the bond geometry information and bounds, lower/upper bounds
        for conjugated terminal bonds are symmetrized.
    """

    def _sort_bond_ids(bond_ids: tuple[tuple | list]) -> tuple[tuple, ...]:
        return tuple(tuple(sorted(_)) for _ in bond_ids)

    def _get_terminal_group_matches(_mol: Mol) -> tuple[tuple, ...]:
        qsmarts = "[O,N;D1;$([O,N;D1]-[*]=[O,N;D1]),$([O,N;D1]=[*]-[O,N;D1])]~[*]"
        qsmarts = MolFromSmarts(qsmarts)
        matches = _mol.GetSubstructMatches(qsmarts)
        return _sort_bond_ids(matches)

    # sorting the atom types to use them as an index
    df["atom_types_sorted"] = df["atom_types"].apply(lambda a: tuple(sorted(a.split("--"))))
    # conjugated terminal atoms matches
    matches = _get_terminal_group_matches(mol)
    matched = df[df["atom_pair"].isin(matches)].copy()
    # min and max of lower and upper bounds
    grouped = matched.groupby("atom_types_sorted").agg({"lower_bound": np.amin, "upper_bound": np.amax})
    # updating the matches dataframe and the original dataframe
    index_orig = matched.index
    matched = matched.set_index("atom_types_sorted")
    matched.update(grouped)
    matched = matched.set_index(index_orig)
    df.update(matched)
    return df.drop(columns=["atom_types_sorted"])


def check_geometry(  # noqa: PLR0913, PLR0915
    mol_pred: Mol,
    threshold_bad_bond_length: float = 0.2,
    threshold_clash: float = 0.2,
    threshold_bad_angle: float = 0.2,
    bound_matrix_params: dict[str, Any] = bound_matrix_params,
    ignore_hydrogens: bool = True,
    sanitize: bool = True,
    symmetrize_conjugated_terminal_groups: bool = True,
) -> dict[str, Any]:
    """Use RDKit distance geometry bounds to check the geometry of a molecule.

    Args:
        mol_pred: Predicted molecule (docked ligand). Only the first conformer will be checked.
        threshold_bad_bond_length: Bond length threshold in relative percentage. 0.2 means that bonds may be up to 20%
            longer than DG bounds. Defaults to 0.2.
        threshold_clash: Threshold for how much overlap constitutes a clash. 0.2 means that the two atoms may be up to
            80% of the lower bound apart. Defaults to 0.2.
        threshold_bad_angle: Bond angle threshold in relative percentage. 0.2 means that bonds may be up to 20%
            longer than DG bounds. Defaults to 0.2.
        bound_matrix_params: Parameters passe to RDKit's GetMoleculeBoundsMatrix function.
        ignore_hydrogens: Whether to ignore hydrogens. Defaults to True.
        sanitize: Sanitize molecule before running DG module (recommended). Defaults to True.
        symmetrize_conjugated_terminal_groups: Will symmetrize the lower and upper bounds of the terminal
            conjugated bonds. Defaults to True.

    Returns:
        PoseBusters results dictionary.
    """
    mol = deepcopy(mol_pred)
    results = _empty_results.copy()

    if mol.GetNumConformers() == 0:
        logger.warning("Molecule does not have a conformer.")
        return {"results": results}

    if mol.GetNumAtoms() == 1:
        logger.warning(f"Molecule has only {mol.GetNumAtoms()} atoms.")
        results[col_angles_result] = True
        results[col_bonds_result] = True
        results[col_clash_result] = True
        return {"results": results}

    # sanitize to ensure DG works or manually process molecule
    try:
        if sanitize:
            flags = SanitizeMol(mol)
            assert flags == 0, f"Sanitization failed with flags {flags}"
    except Exception:
        return {"results": results}

    # get bonds and angles
    bond_set = sorted(_get_bond_atom_indices(mol))  # tuples
    angles = sorted(_get_angle_atom_indices(bond_set))  # triples
    angle_set = {(a[0], a[2]): a for a in angles}  # {tuples : triples}

    if len(bond_set) == 0:
        logger.warning("Molecule has no bonds.")

    # distance geometry bounds, lower triangle min distances, upper triangle max distances
    try:
        bounds = GetMoleculeBoundsMatrix(mol, **bound_matrix_params)
    except RuntimeError as e:
        logger.warning(f"RDKit distance geometry failed with error: {e}")
        return {"results": results}

    # indices
    lower_triangle_idcs = np.tril_indices(mol.GetNumAtoms(), k=-1)
    upper_triangle_idcs = (lower_triangle_idcs[1], lower_triangle_idcs[0])

    # 1,2- distances
    df_12 = pd.DataFrame()
    df_12["atom_pair"] = list(zip(*upper_triangle_idcs))  # indices have i1 < i2
    df_12["atom_types"] = [
        "--".join(tuple(mol.GetAtomWithIdx(int(j)).GetSymbol() for j in i)) for i in df_12["atom_pair"]
    ]
    df_12["angle"] = df_12["atom_pair"].apply(lambda x: angle_set.get(x, None))
    df_12["has_hydrogen"] = [_has_hydrogen(mol, i) for i in df_12["atom_pair"]]
    df_12["is_bond"] = [i in bond_set for i in df_12["atom_pair"]]
    df_12["is_angle"] = df_12["angle"].apply(lambda x: x is not None)
    df_12[col_lb] = bounds[lower_triangle_idcs]
    df_12[col_ub] = bounds[upper_triangle_idcs]

    if symmetrize_conjugated_terminal_groups:
        df_12 = symmetrize_conjugated_terminal_bonds(df_12, mol)

    # add observed dimensions
    conformer = mol.GetConformer()
    conf_distances = _pairwise_distance(conformer.GetPositions())
    df_12["conf_id"] = conformer.GetId()
    df_12["distance"] = conf_distances[lower_triangle_idcs]

    # set types
    df_12 = df_12.astype(
        {
            "has_hydrogen": "bool",
            "is_bond": "bool",
            "is_angle": "bool",
            col_lb: "float",
            col_ub: "float",
            "conf_id": "int",
            "distance": "float",
        }
    )

    # remove hydrogens if requested
    if ignore_hydrogens:
        df_12 = df_12.loc[df_12["has_hydrogen"].eq(False), :]

    # calculate violations
    df_bonds = _bond_check(df_12)
    df_clash = _clash_check(df_12)
    df_angles = _angle_check(df_12)

    # bond statistics
    results[col_n_bonds] = len(df_bonds)
    results[col_n_short_bonds] = sum(df_bonds[col_pe] < -threshold_bad_bond_length)
    results[col_n_long_bonds] = sum(df_bonds[col_pe] > threshold_bad_bond_length)
    results[col_n_good_bonds] = results[col_n_bonds] - results[col_n_short_bonds] - results[col_n_long_bonds]
    results[col_bonds_result] = results[col_n_good_bonds] == results[col_n_bonds]
    results[col_shortest_bond] = (df_bonds["distance"] / df_bonds["lower_bound"]).min()
    results[col_longest_bond] = (df_bonds["distance"] / df_bonds["upper_bound"]).max()

    # angle statistics
    results[col_n_angles] = len(df_angles)
    results[col_n_bad_angles] = sum(df_angles[col_bape] > threshold_bad_angle)
    results[col_n_good_angles] = results[col_n_angles] - results[col_n_bad_angles]
    results[col_angles_result] = results[col_n_good_angles] == results[col_n_angles]
    lb_extreme_angle = 2 - (df_angles["distance"] / df_angles["lower_bound"]).min()
    ub_extreme_angle = (df_angles["distance"] / df_angles["upper_bound"]).max()
    results[col_extremest_angle] = max(lb_extreme_angle, ub_extreme_angle)

    # steric clash statistics
    results[col_n_noncov] = len(df_clash)
    results[col_n_clashes] = sum(df_clash[col_bpe] < -threshold_clash)
    results[col_n_good_noncov] = results[col_n_noncov] - results[col_n_clashes]
    results[col_clash_result] = results[col_n_good_noncov] == results[col_n_noncov]
    results[col_closest_noncov] = (df_clash["distance"] / df_clash["lower_bound"]).min()

    return {"results": results, "details": {"bonds": df_bonds, "clash": df_clash, "angles": df_angles}}


def _bond_check(df: pd.DataFrame) -> pd.DataFrame:
    # bonds can be too short or too long
    df = df[df["is_bond"]].copy()
    df[col_pe] = df.apply(lambda x: bpe(*x[["distance", col_lb, col_ub]]), axis=1)
    return df


def _angle_check(df: pd.DataFrame) -> pd.DataFrame:
    # angles have no direction (we do not know if larger or bigger beyond bounds)
    df = df[(~df["is_bond"]) & (df["is_angle"])].copy()
    df[col_bape] = df.apply(lambda x: bape(*x[["distance", col_lb, col_ub]]), axis=1)
    return df


def _clash_check(df: pd.DataFrame) -> pd.DataFrame:
    # clash is only when lower bound is violated
    df = df[(~df["is_bond"]) & (~df["is_angle"])].copy()

    def _lb_pe(value: float, lower_bound: float) -> float:
        if value >= lower_bound:
            return 0.0
        return pe(value, lower_bound)

    df[col_bpe] = df.apply(lambda x: _lb_pe(*x[["distance", col_lb]]), axis=1)
    return df


def _get_bond_atom_indices(mol: Mol) -> list[tuple[int, int]]:
    bonds = []
    for bond in mol.GetBonds():
        bond_tuple = (bond.GetBeginAtomIdx(), bond.GetEndAtomIdx())
        bond_tuple = _sort_bond(bond_tuple)
        bonds.append(bond_tuple)
    return bonds


def _get_angle_atom_indices(bonds: list[tuple[int, int]]) -> list[tuple[int, int, int]]:
    """Check all combinations of bonds to generate list of molecule angles."""
    angles = []
    bonds = list(bonds)
    for i in range(len(bonds)):
        for j in range(i + 1, len(bonds)):
            angle = _two_bonds_to_angle(bonds[i], bonds[j])
            if angle is not None:
                angles.append(angle)
    return angles


def _two_bonds_to_angle(bond1: tuple[int, int], bond2: tuple[int, int]) -> None | tuple[int, int, int]:
    set1 = set(bond1)
    set2 = set(bond2)
    all_atoms = set1 | set2
    # angle requires two bonds to share exactly one atom, that is we must have 3 atoms
    if len(all_atoms) != 3:  # noqa: PLR2004
        return None
    # find shared atom
    shared_atom = set1 & set2
    other_atoms = all_atoms - shared_atom
    return (min(other_atoms), shared_atom.pop(), max(other_atoms))


def _sort_bond(bond: tuple[int, int]) -> tuple[int, int]:
    return (min(bond), max(bond))


def _pairwise_distance(x: np.ndarray) -> np.ndarray:
    return np.linalg.norm(x[:, None, :] - x[None, :, :], axis=-1)


def _has_hydrogen(mol: Mol, idcs: Iterable[int]) -> bool:
    return any(_is_hydrogen(mol, idx) for idx in idcs)


def _is_hydrogen(mol: Mol, idx: int) -> bool:
    return mol.GetAtomWithIdx(int(idx)).GetAtomicNum() == 1



from __future__ import annotations

import pytest
from rdkit.Chem.rdmolops import AddHs, RemoveHs

from posebusters.modules.sucos import check_sucos


def test_check_sucos(mol_rq3_x00, mol_rq3_x01, mol_rq3_x10, mol_small_7brv_f5r_7wb6_f5r, mol_large_7brv_f5r_7wb6_f5r):
    out = check_sucos(mol_rq3_x00, mol_rq3_x00)
    assert out["results"]["sucos"] == pytest.approx(1.0, abs=1e-6)

    out = check_sucos(mol_rq3_x00, mol_rq3_x01)
    assert out["results"]["sucos"] == pytest.approx(0.5, abs=0.1)

    out = check_sucos(mol_rq3_x00, mol_rq3_x10)
    assert out["results"]["sucos"] == pytest.approx(0.0, abs=1e-6)


def test_check_sucos_hydrogens(mol_small_7brv_f5r_7wb6_f5r, mol_large_7brv_f5r_7wb6_f5r):
    out = check_sucos(
        AddHs(mol_large_7brv_f5r_7wb6_f5r, addCoords=True), AddHs(mol_small_7brv_f5r_7wb6_f5r, addCoords=True)
    )
    assert out["results"]["sucos"] <= 1.0

    out = check_sucos(RemoveHs(mol_large_7brv_f5r_7wb6_f5r), RemoveHs(mol_small_7brv_f5r_7wb6_f5r))
    assert out["results"]["sucos"] <= 1.0

    out = check_sucos(AddHs(mol_large_7brv_f5r_7wb6_f5r, addCoords=True), RemoveHs(mol_small_7brv_f5r_7wb6_f5r))
    assert out["results"]["sucos"] <= 1.0

    out = check_sucos(RemoveHs(mol_small_7brv_f5r_7wb6_f5r), AddHs(mol_large_7brv_f5r_7wb6_f5r, addCoords=True))
    assert out["results"]["sucos"] <= 1.0


def test_check_sucos_multiple_ground_truth(mol_one_true_1w1p, mol_true_1w1p):
    out = check_sucos(mol_one_true_1w1p, mol_true_1w1p)
    assert out["results"]["sucos"] == pytest.approx(1.0, abs=1e-6)


def test_get_sucos_score_065(mol_065, mol_065_left, mol_065_right):
    out = check_sucos(mol_065, mol_065)
    assert out["results"]["sucos"] == pytest.approx(1.0, abs=1e-6)

    out = check_sucos(mol_065_left, mol_065_right)
    assert out["results"]["sucos"] == pytest.approx(0.0, abs=1e-6)


def test_get_sucos_score_TMO(mol_TMO):
    # this molecule has no features
    out = check_sucos(mol_TMO, mol_TMO)
    assert out["results"]["sucos"] == 1.0


def test_get_sucos_score_2YU_HQT(mol_2YU, mol_HQT):
    out = check_sucos(mol_2YU, mol_HQT)
    assert out["results"]["sucos"] <= 1.0



"""Module to check loading of protein and ligand files."""

from __future__ import annotations

from typing import Any

from rdkit.Chem.rdchem import Mol


def check_loading(mol_pred: Any = None, mol_true: Any = None, mol_cond: Any = None) -> dict[str, dict[str, bool]]:
    """Check that molecule files were loaded successfully.

    Args:
        mol_pred: Predicted molecule. Defaults to None.
        mol_true: Ground truth molecule. Defaults to None.
        mol_cond: Conditioning molecule. Defaults to None.

    Returns:
        PoseBusters results dictionary.
    """
    results = {
        "mol_pred_loaded": isinstance(mol_pred, Mol),
        "mol_true_loaded": isinstance(mol_true, Mol),
        "mol_cond_loaded": isinstance(mol_cond, Mol),
    }
    return {"results": results}



"""Module to check intermolecular distances between ligand and protein."""

from __future__ import annotations

from typing import Any

import numpy as np
import pandas as pd
from rdkit.Chem.rdchem import GetPeriodicTable, Mol

from ..tools.protein import get_atom_type_mask

_periodic_table = GetPeriodicTable()


def check_intermolecular_distance(  # noqa: PLR0913
    mol_pred: Mol,
    mol_cond: Mol,
    radius_type: str = "vdw",
    radius_scale: float = 1.0,
    clash_cutoff: float = 0.75,
    ignore_types: set[str] = {"hydrogens"},
    max_distance: float = 5.0,
    search_distance: float = 6.0,
) -> dict[str, Any]:
    """Check that predicted molecule is not too close and not too far away from conditioning molecule.

    Args:
        mol_pred: Predicted molecule (docked ligand) with one conformer.
        mol_cond: Conditioning molecule (protein) with one conformer.
        radius_type: Type of atomic radius to use. Possible values are "vdw" (van der Waals) and "covalent".
            Defaults to "vdw".
        radius_scale: Scaling factor for the atomic radii. Defaults to 0.8.
        clash_cutoff: Threshold for how much the atoms may overlap before a clash is reported. Defaults
            to 0.05.
        ignore_types: Which types of atoms to ignore in mol_cond. Possible values to include are "hydrogens", "protein",
            "organic_cofactors", "inorganic_cofactors", "waters". Defaults to {"hydrogens"}.
        max_distance: Maximum distance (in Angstrom) predicted and conditioning molecule may be apart to be considered
            as valid. Defaults to 5.0.

    Returns:
        PoseBusters results dictionary.
    """
    coords_ligand = mol_pred.GetConformer().GetPositions()
    coords_protein = mol_cond.GetConformer().GetPositions()

    atoms_ligand = np.array([a.GetSymbol() for a in mol_pred.GetAtoms()])
    atoms_protein_all = np.array([a.GetSymbol() for a in mol_cond.GetAtoms()])

    idxs_ligand = np.array([a.GetIdx() for a in mol_pred.GetAtoms()])
    idxs_protein = np.array([a.GetIdx() for a in mol_cond.GetAtoms()])

    mask = [a.GetSymbol() != "H" for a in mol_pred.GetAtoms()]
    coords_ligand = coords_ligand[mask, :]
    atoms_ligand = atoms_ligand[mask]
    mask_ligand_idxs = idxs_ligand[mask]
    if ignore_types:
        mask = get_atom_type_mask(mol_cond, ignore_types)
        coords_protein = coords_protein[mask, :]
        atoms_protein_all = atoms_protein_all[mask]
        mask_protein_idxs = idxs_protein[mask]

    # get radii
    radius_ligand = _get_radii(atoms_ligand, radius_type)
    radius_protein_all = _get_radii(atoms_protein_all, radius_type)

    # select atoms that are close to ligand to check for clash
    distances_all = _pairwise_distance(coords_ligand, coords_protein)
    mask_protein = distances_all.min(axis=0) <= search_distance
    distances = distances_all[:, mask_protein]
    radius_protein = radius_protein_all[mask_protein]
    atoms_protein = atoms_protein_all[mask_protein]
    mask_protein_idxs = mask_protein_idxs[mask_protein]

    radius_sum = radius_ligand[:, None] + radius_protein[None, :]
    relative_distance = distances / radius_sum
    violations = relative_distance < 1 / radius_scale

    if distances.size > 0:
        violations[np.unravel_index(distances.argmin(), distances.shape)] = True  # add smallest distances as info
        violations[np.unravel_index(relative_distance.argmin(), relative_distance.shape)] = True
    violation_ligand, violation_protein = np.where(violations)
    reverse_ligand_idxs = mask_ligand_idxs[violation_ligand]
    reverse_protein_idxs = mask_protein_idxs[violation_protein]

    # collect details around those violations in a dataframe
    details = pd.DataFrame()
    details["ligand_atom_id"] = reverse_ligand_idxs
    details["protein_atom_id"] = reverse_protein_idxs
    details["ligand_element"] = [atoms_ligand[i] for i in violation_ligand]
    details["protein_element"] = [atoms_protein[i] for i in violation_protein]
    details["ligand_vdw"] = [radius_ligand[i] for i in violation_ligand]
    details["protein_vdw"] = [radius_protein[i] for i in violation_protein]
    details["sum_radii"] = details["ligand_vdw"] + details["protein_vdw"]
    details["distance"] = distances[violation_ligand, violation_protein]
    details["sum_radii_scaled"] = details["sum_radii"] * radius_scale
    details["relative_distance"] = details["distance"] / details["sum_radii_scaled"]
    details["clash"] = details["relative_distance"] < clash_cutoff

    results = {
        "smallest_distance": details["distance"].min(),
        "not_too_far_away": details["distance"].min() <= max_distance,
        "num_pairwise_clashes": details["clash"].sum(),
        "no_clashes": not details["clash"].any(),
    }

    # add most extreme values to results table
    i = np.argmin(details["relative_distance"]) if len(details) > 0 else None
    most_extreme = {"most_extreme_" + c: details.loc[i][str(c)] if i is not None else pd.NA for c in details.columns}
    results = {**results, **most_extreme}

    return {"results": results, "details": details}


def _pairwise_distance(x: np.ndarray, y: np.ndarray) -> np.ndarray:
    return np.linalg.norm(x[:, None, :] - y[None, :, :], axis=-1)


def _get_radii(atoms: np.ndarray, radius_type: str) -> np.ndarray:
    if radius_type == "vdw":
        return np.array([_periodic_table.GetRvdw(a) for a in atoms])
    elif radius_type == "covalent":
        return np.array([_periodic_table.GetRcovalent(a) for a in atoms])
    else:
        raise ValueError(f"Unknown radius type {radius_type}. Valid values are 'vdw' and 'covalent'.")



"""Module to check chemistry of molecules."""

from __future__ import annotations

from copy import deepcopy
from typing import Any

import pandas as pd
from rdkit import RDLogger
from rdkit.Chem.inchi import InchiReadWriteError
from rdkit.Chem.rdchem import Mol
from rdkit.Chem.rdmolops import DetectChemistryProblems, GetMolFrags, SanitizeFlags, SanitizeMol

from ..tools.inchi import get_inchi


def check_chemistry_using_rdkit(mol_pred: Mol) -> dict[str, dict[str, bool] | dict[str, bool | str] | pd.DataFrame]:
    """Check sanity of molecule using RDKit sanitization rules."""
    assert isinstance(mol_pred, Mol)
    mol = deepcopy(mol_pred)

    RDLogger.DisableLog("rdApp.*")  # type: ignore[attr-defined]
    errors = DetectChemistryProblems(mol, sanitizeOps=SanitizeFlags.SANITIZE_ALL)
    RDLogger.EnableLog("rdApp.*")  # type: ignore[attr-defined]

    messages = [error.Message() for error in errors]
    atom_indices = [e.GetAtomIndices() if hasattr(e, "GetAtomIndices") else e.GetAtomIdx() for e in errors]
    types = [error.GetType() for error in errors]

    details = pd.DataFrame(dict(messages=messages, atom_indices=atom_indices, types=types))

    results = {
        "passes_valence_checks": "AtomValenceException" not in types,
        "passes_kekulization": "AtomKekulizeException" not in types,
        "passes_rdkit_sanity_checks": len(errors) == 0,
    }
    return {"results": results, "details": details}


def check_chemistry_using_inchi(mol_pred: Mol) -> dict[str, dict[str, str | bool | None]]:
    """Check sanity of a molecule using InChI rules."""
    assert isinstance(mol_pred, Mol)
    mol = deepcopy(mol_pred)

    passes = False

    try:
        inchi = get_inchi(mol, inchi_strict=False)
        passes = True
        errors = ""
    except InchiReadWriteError as e:
        errors = str(e.args[1])
        inchi = None
    except Exception as e:
        errors = str(e)
        inchi = None

    return {"results": {"inchi_convertible": passes}, "details": dict(errors=errors, inchi=inchi)}


def check_all_atoms_connected(mol_pred: Mol) -> dict[str, dict[str, bool] | dict[str, int]]:
    """Check if all atoms in a molecule are connected."""
    assert isinstance(mol_pred, Mol)
    mol = deepcopy(mol_pred)

    num_frags = len(GetMolFrags(mol, asMols=False, sanitizeFrags=False))
    results = {"all_atoms_connected": num_frags <= 1}
    details = {"number_fragments": num_frags}

    return {"results": results, "details": details}


def check_chemistry(mol_pred: Mol) -> dict[str, Any]:
    """Check sanity using RDKit and connectedness.

    Notes:
    - Only for backward compatibility. Use `check_chemistry_using_rdkit` and `check_all_atoms_connected` separately.
    """

    rdkit_sanity = check_chemistry_using_rdkit(mol_pred)
    connectedness = check_all_atoms_connected(mol_pred)
    return {
        "results": rdkit_sanity["results"] | connectedness["results"],
        "details": rdkit_sanity["details"],
    }


def check_radicals(mol_pred: Mol) -> dict[str, Any]:
    """Check that a molecule has no radicals."""

    radicals_before_sanitization = any(a.GetNumRadicalElectrons() for a in mol_pred.GetAtoms())
    SanitizeMol(mol_pred, catchErrors=True)
    radicals_after_sanitization = any(a.GetNumRadicalElectrons() for a in mol_pred.GetAtoms())

    return {
        "results": {
            "no_radicals": not radicals_after_sanitization,
            "no_radicals_before_sanitization": not radicals_before_sanitization,
        }
    }



"""Load datasets from the datasets folder."""

from pathlib import Path

import pandas as pd

dataset_dir = Path(__file__).parent


def _prepend_path(file_path: Path | str) -> Path:
    return dataset_dir / file_path


def _load_dataset(cols: list[str]) -> pd.DataFrame:
    files = pd.read_csv(dataset_dir / "pdb.csv")
    files[cols] = files[cols].apply(_prepend_path)
    return files[["name"] + cols]


def four_redocks() -> pd.DataFrame:
    """Load the four docks dataset."""
    return _load_dataset(["mol_pred", "mol_true", "mol_cond"])


def four_docks() -> pd.DataFrame:
    """Load the four docks dataset."""
    return _load_dataset(["mol_pred", "mol_cond"])


def four_mols() -> pd.DataFrame:
    """Load the four mols dataset."""
    return _load_dataset(["mol_pred"])



"""Module to check volume overlap between docked ligand and protein."""

from __future__ import annotations

import logging

import numpy as np
from rdkit.Chem.rdchem import Mol
from rdkit.Chem.rdShapeHelpers import ShapeTverskyIndex

from ..tools.molecules import delete_atoms
from ..tools.protein import get_atom_type_mask

logger = logging.getLogger(__name__)


def check_volume_overlap(  # noqa: PLR0913
    mol_pred: Mol,
    mol_cond: Mol,
    clash_cutoff: float = 0.05,
    vdw_scale: float = 0.8,
    ignore_types: set[str] = {"hydrogens"},
    search_distance: float = 6.0,
) -> dict[str, dict[str, float | bool]]:
    """Check volume overlap between ligand and protein.

    Args:
        mol_pred: Predicted molecule (docked ligand) with one conformer.
        mol_cond: Conditioning molecule (protein) with one conformer.
        clash_cutoff: Threshold for how much volume overlap is allowed. This is the maximum share of volume of
            `mol_pred` allowed to overlap with `mol_cond`. Defaults to 0.05.
        vdw_scale: Scaling factor for the van der Waals radii which define the volume around each atom. Defaults to 0.8.
        ignore_types: Which types of atoms in mol_cond to ignore. Possible values to include are "hydrogens", "protein",
            "organic_cofactors", "inorganic_cofactors", "waters". Defaults to {"hydrogens"}.

    Returns:
        PoseBusters results dictionary.
    """
    assert isinstance(mol_pred, Mol)
    assert isinstance(mol_cond, Mol)

    # filter by atom types
    keep_mask = np.array(get_atom_type_mask(mol_cond, ignore_types))
    mol_cond = _filter_by_mask(mol_cond, keep_mask)
    if mol_cond.GetNumAtoms() == 0:
        return {"results": {"volume_overlap": np.nan, "no_volume_clash": True}}

    # filter by distance --> this is slowing this function down
    distances = _pairwise_distance(mol_pred.GetConformer().GetPositions(), mol_cond.GetConformer().GetPositions())
    keep_mask = distances.min(axis=0) <= search_distance * vdw_scale
    mol_cond = _filter_by_mask(mol_cond, keep_mask)
    if mol_cond.GetNumAtoms() == 0:
        return {"results": {"volume_overlap": np.nan, "no_volume_clash": True}}

    ignore_hydrogens = "hydrogens" in ignore_types
    overlap = ShapeTverskyIndex(mol_pred, mol_cond, alpha=1, beta=0, vdwScale=vdw_scale, ignoreHs=ignore_hydrogens)

    results = {
        "volume_overlap": overlap,
        "no_volume_clash": overlap <= clash_cutoff,
    }

    return {"results": results}


def _pairwise_distance(x: np.ndarray, y: np.ndarray) -> np.ndarray:
    return np.linalg.norm(x[:, None, :] - y[None, :, :], axis=-1)


def _filter_by_mask(mol: Mol, mask: np.ndarray) -> Mol:
    if mask.sum() < len(mask):
        mol = delete_atoms(mol, np.where(~mask)[0].tolist())  # type: ignore[arg-type]
    return mol



from __future__ import annotations

import math

import numpy as np

from posebusters.modules.energy_ratio import check_energy_ratio


def test_check_energy_ratio(mol_pm2):
    # molecule has valid conformation
    out = check_energy_ratio(mol_pm2)
    assert math.isfinite(out["results"]["mol_pred_energy"])
    assert math.isfinite(out["results"]["ensemble_avg_energy"])
    assert math.isfinite(out["results"]["energy_ratio"])
    assert out["results"]["energy_ratio_passes"]


def test_check_energy_ratio_14gs_0(mol_pred_14gs_gen0):
    # molecule is chemically insane, but conformation looks alright
    out = check_energy_ratio(mol_pred_14gs_gen0, threshold_energy_ratio=10.0)
    assert math.isfinite(out["results"]["mol_pred_energy"])
    assert math.isfinite(out["results"]["ensemble_avg_energy"])
    assert math.isfinite(out["results"]["energy_ratio"])
    assert out["results"]["energy_ratio_passes"]


def test_check_energy_ratio_1afs_87(mol_pred_1afs_gen87):
    # molecule can be sanitized and hydrogens can be added, conformation not horrible
    out = check_energy_ratio(mol_pred_1afs_gen87)
    assert math.isfinite(out["results"]["mol_pred_energy"])
    assert math.isfinite(out["results"]["ensemble_avg_energy"])
    assert math.isfinite(out["results"]["energy_ratio"])
    assert out["results"]["energy_ratio_passes"]


def test_check_energy_ratio_1afs_94(mol_pred_1afs_gen94):
    # molecule is chemically insane, but conformation looks alright
    out = check_energy_ratio(mol_pred_1afs_gen94)
    assert math.isfinite(out["results"]["mol_pred_energy"])
    assert math.isfinite(out["results"]["ensemble_avg_energy"])
    assert math.isfinite(out["results"]["energy_ratio"])
    assert out["results"]["energy_ratio_passes"]


def test_check_energy_ratio_1jn2_3(mol_pred_1jn2_gen3):
    # molecule can be sanitized and hydrogens can be added, conformation not horrible
    out = check_energy_ratio(mol_pred_1jn2_gen3)
    assert math.isfinite(out["results"]["mol_pred_energy"])
    assert math.isfinite(out["results"]["ensemble_avg_energy"])
    assert math.isfinite(out["results"]["energy_ratio"])
    assert out["results"]["energy_ratio_passes"]


def test_check_energy_ratio_1jn2_62(mol_pred_1jn2_gen62):
    # molecule is chemically insane, but conformation looks alright
    out = check_energy_ratio(mol_pred_1jn2_gen62)
    assert math.isfinite(out["results"]["mol_pred_energy"])
    assert math.isfinite(out["results"]["ensemble_avg_energy"])
    assert math.isfinite(out["results"]["energy_ratio"])
    assert out["results"]["energy_ratio_passes"]


def test_check_energy_ratio_approximate_consistency(mol_1a30_clash_2, mol_1a30_clash_3):
    energy_2 = check_energy_ratio(mol_1a30_clash_2)["results"]["mol_pred_energy"]
    energy_3 = check_energy_ratio(mol_1a30_clash_3)["results"]["mol_pred_energy"]

    # check numbers are approximately equal
    assert np.isclose(energy_2, energy_3)


def test_check_energy_ratio_disconnected_atoms(mol_disconnnected_atoms):
    # no bonds, just disconnected atoms -> energy is 0
    out = check_energy_ratio(mol_disconnnected_atoms)
    assert math.isfinite(out["results"]["mol_pred_energy"])
    assert math.isfinite(out["results"]["ensemble_avg_energy"])
    assert math.isfinite(out["results"]["energy_ratio"])
    assert out["results"]["energy_ratio_passes"] is True


def test_check_energy_ratio_iron_ion(mol_fe2plus):
    # test fails because UFF does not have parameters for Fe 2+ ion
    out = check_energy_ratio(mol_fe2plus)
    assert math.isnan(out["results"]["energy_ratio_passes"])



from __future__ import annotations

from rdkit.Chem.rdmolfiles import MolFromSmiles

from posebusters.modules.sanity import check_chemistry_using_inchi, check_chemistry_using_rdkit


def test_check_rdkit_sanity():
    mol_pass = MolFromSmiles("C=C", sanitize=True)
    assert check_chemistry_using_rdkit(mol_pass)["results"]["passes_rdkit_sanity_checks"] is True

    mol_pass = MolFromSmiles("c1ccccc1", sanitize=True)
    assert check_chemistry_using_rdkit(mol_pass)["results"]["passes_rdkit_sanity_checks"] is True

    mol_pass = MolFromSmiles("CCO", sanitize=True)
    assert check_chemistry_using_rdkit(mol_pass)["results"]["passes_rdkit_sanity_checks"] is True

    mol_fail = MolFromSmiles("F=F", sanitize=False)
    assert check_chemistry_using_rdkit(mol_fail)["results"]["passes_rdkit_sanity_checks"] is False
    assert check_chemistry_using_rdkit(mol_fail)["results"]["passes_valence_checks"] is False
    assert check_chemistry_using_rdkit(mol_fail)["results"]["passes_kekulization"] is True

    mol_fail = MolFromSmiles("c1ccccc1c", sanitize=False)
    assert check_chemistry_using_rdkit(mol_fail)["results"]["passes_rdkit_sanity_checks"] is False
    assert check_chemistry_using_rdkit(mol_fail)["results"]["passes_valence_checks"] is True
    assert check_chemistry_using_rdkit(mol_fail)["results"]["passes_kekulization"] is False


def test_check_chemistry_using_inchi_1afs(mol_pred_1afs_gen87):
    info = check_chemistry_using_inchi(mol_pred_1afs_gen87)

    assert info["results"]["inchi_convertible"] is False


def test_check_chemistry_using_inchi_1afs_94(mol_pred_1afs_gen94):
    info = check_chemistry_using_inchi(mol_pred_1afs_gen94)

    assert info["results"]["inchi_convertible"] is True


def test_check_chemistry_using_inchi_1jn2_3(mol_pred_1jn2_gen3):
    info = check_chemistry_using_inchi(mol_pred_1jn2_gen3)

    assert info["results"]["inchi_convertible"] is False


def test_check_chemistry_using_inchi_1jn2_62(mol_pred_1jn2_gen62):
    info = check_chemistry_using_inchi(mol_pred_1jn2_gen62)

    assert info["results"]["inchi_convertible"] is True



from __future__ import annotations

import logging
import re
from copy import deepcopy
from re import findall

from rdkit.Chem.inchi import MolFromInchi, MolToInchi
from rdkit.Chem.MolStandardize.rdMolStandardize import Uncharger
from rdkit.Chem.rdchem import AtomValenceException, Mol
from rdkit.Chem.rdmolops import AssignStereochemistryFrom3D, RemoveHs, RemoveStereochemistry, SanitizeMol

from .logging import CaptureLogger
from .molecules import add_stereo_hydrogens, neutralize_atoms, remove_isotopic_info

logger = logging.getLogger(__name__)


def get_inchi(mol: Mol, inchi_strict: bool = False) -> str:
    """Get inchi of a molecule."""

    with CaptureLogger() as log:
        SanitizeMol(mol)
        inchi = str(MolToInchi(mol, treatWarningAsError=inchi_strict))
        # check inchi because inchi generation does not raise an error if the inchi is invalid
        if MolFromInchi(inchi, sanitize=True) is None:
            warnings = re.split(r"\[.+?\] WARNING: ", log.get("WARNING", ""))[1:]
            errors = re.split(r"\[.+?\] ERROR: ", log.get("ERROR", ""))[1:]
            message = f"InChI ERRORs: {'; '.join(errors)} WARNINGs: {'; '.join(warnings)}"
            raise Exception(message)
    return inchi


def standardize_and_get_inchi(mol: Mol, options: str = "", log_level=None, warnings_as_errors=False) -> str:
    """Return InChI after standardising molecule and inferring stereo from coordinates."""

    mol = deepcopy(mol)

    if flags := SanitizeMol(mol, catchErrors=True):
        logger.debug("Cannot get InChI because molecule doesn't sanitize due to %s.", flags)
        return ""

    # standardise molecule
    mol = remove_isotopic_info(mol)

    # assign stereo from 3D coordinates only if 3D coordinates are present
    has_pose = mol.GetNumConformers() > 0
    if has_pose:
        RemoveStereochemistry(mol)

    mol = RemoveHs(mol)
    try:
        mol = neutralize_atoms(mol)
    except AtomValenceException:
        logger.warning("Failed to neutralize molecule. Using uncharger. InChI check might fail.")
        mol = Uncharger().uncharge(mol)
    mol = add_stereo_hydrogens(mol)

    if has_pose:
        AssignStereochemistryFrom3D(mol, replaceExistingTags=True)

    with CaptureLogger():
        inchi = str(MolToInchi(mol, options=options, logLevel=log_level, treatWarningAsError=warnings_as_errors))

    return inchi


def is_valid_inchi(inchi: str) -> bool:
    """Check that InChI can be parsed and sanitization does not fail."""
    try:
        mol = MolFromInchi(inchi)
        assert mol is not None, "Molecule is None."
        assert not SanitizeMol(mol, catchErrors=True), "Molecule does not sanitize."
        return True
    except Exception:
        return False


def split_inchi(inchi: str) -> dict[str, str]:
    """Split the standard InChI string without isotopic info into its layers."""
    if not inchi.startswith("InChI="):
        raise ValueError("InChI string must start with 'InChI='")

    # inchi always InChi=1S/...formula.../...layer.../...layer.../...layer...
    version = inchi[6:].split(r"/", 1)[0]
    layers = findall(r"(?=.*)(\/[a-z]{0,1})(.*?)(?=\/.*|$)", inchi[6:])

    inchi_parts = {"=": version}
    for prefix, layer in layers:
        # standard inchi strings without isotopic info have each layer no more than once
        assert prefix not in inchi_parts, f"Layer {prefix} more than once!"
        inchi_parts[prefix] = layer

    return inchi_parts



"""Configuration file for the Sphinx documentation builder."""

# -- Project information -----------------------------------------------------

project = "PoseBusters"
copyright = "2023, Martin Buttenschoen"
author = "Martin Buttenschoen"

# get version and release from library. ref https://peps.python.org/pep-0440/
version, release = "major.minor.micro", "major.minor"
with open("../../posebusters/__init__.py") as init_file:
    for line in init_file:
        if line.startswith("__version__"):
            version = line.split("=")[1].strip().strip('"')
            release = ".".join(version.split(".", 2)[:2])
            break

# -- General configuration ---------------------------------------------------

extensions = [
    "hoverxref.extension",
    "sphinx.ext.duration",
    "sphinx.ext.doctest",
    "sphinx.ext.autodoc",
    "sphinx.ext.autosummary",
    # "sphinx.ext.autosectionlabel",
    "sphinx.ext.intersphinx",
    "sphinx.ext.coverage",
    "sphinx.ext.napoleon",
    "sphinx.ext.todo",
    "nbsphinx",
    "sphinxcontrib.programoutput",
]

intersphinx_mapping = {
    "python": ("https://docs.python.org/3/", None),
    "sphinx": ("https://www.sphinx-doc.org/en/master/", None),
}
intersphinx_disabled_domains = ["std"]

templates_path = ["_templates"]
exclude_patterns = ["build", "_build", "Thumbs.db", ".DS_Store"]

language = "en"

# -- Options for HTML output -------------------------------------------------

html_theme = "sphinx_rtd_theme"
html_theme_options = {
    "collapse_navigation": False,
    "display_version": True,
    "prev_next_buttons_location": "both",
    "sticky_navigation": True,
    # 'analytics_id': 'UA-45051049-3',
    # "navigation_depth": 2,
    "includehidden": False,
}
html_static_path = ["_static"]
html_logo = "./_static/logo_square.png"
htmlhelp_basename = "posebusters_doc"
html_last_updated_fmt = "%d %B %Y"

# -- Options for EPUB output -------------------------------------------------

epub_show_urls = "footnote"

# -- Options for todo extension ----------------------------------------------

todo_include_todos = True

# -- Options for napoleon extension ------------------------------------------

# https://www.sphinx-doc.org/en/master/usage/extensions/napoleon.html
napoleon_google_docstring = True
napoleon_numpy_docstring = False
napoleon_include_init_with_doc = False
napoleon_include_private_with_doc = False

# -- Options for coverage extension ------------------------------------------
coverage_show_missing_items = True

# -- Options for hoverxref extension -----------------------------------------
hoverxref_auto_ref = True
hoverxref_role_types = {
    "class": "tooltip",
    "command": "tooltip",
    "confval": "tooltip",
    "hoverxref": "tooltip",
    "mod": "tooltip",
    "ref": "tooltip",
    "reqmeta": "tooltip",
    "setting": "tooltip",
    "signal": "tooltip",
}
hoverxref_roles = ["command", "reqmeta", "setting", "signal"]

# -- Options for programoutput extension -------------------------------------
programoutput_prompt_template = ">>> {command}\n{output}"



"""Provides functions for loading molecules from files."""

from __future__ import annotations

import logging
from collections.abc import Generator, Iterable
from pathlib import Path

from rdkit.Chem.AllChem import AssignBondOrdersFromTemplate
from rdkit.Chem.rdchem import Mol
from rdkit.Chem.rdmolfiles import (
    MolFromMol2File,
    MolFromMolBlock,
    MolFromMolFile,
    MolFromPDBFile,
    MolFromSmiles,
    SDMolSupplier,
)
from rdkit.Chem.rdmolops import AddHs, AssignStereochemistryFrom3D, Cleanup, SanitizeMol
from rdkit.rdBase import LogToPythonLogger

from ..tools.logging import CaptureLogger

LogToPythonLogger()
logger = logging.getLogger(__name__)


def safe_load_mol(path: Path | Mol, load_all: bool = False, **load_params) -> Mol | None:
    """Load one molecule from a file, optionally adding hydrogens and assigning bond orders.

    Args:
        path: Path to file containing molecule.

    Returns:
        Molecule object or None if loading failed.
    """
    if isinstance(path, Mol):
        return path

    path = Path(path)
    if not path.exists():
        logger.warning("File %s does not exist", path)
        return None

    try:
        with CaptureLogger():
            mol = _load_mol(path, load_all=load_all, **load_params)
        return mol
    except Exception as exception:
        logger.warning("Could not load molecule from %s with error: %s", path, exception)
    return None


def safe_supply_mols(
    path: Path, load_all=True, sanitize=True, indices: Iterable[int] | None = None, **load_params
) -> Generator[Mol | None]:
    """Supply molecules from a file, optionally adding hydrogens and assigning bond orders.

    Args:
        path: Path to file containing molecules.

    Returns:
        Molecule object or None if loading failed.
    """
    if isinstance(path, Mol):
        yield path
        return None

    path = Path(path)
    if not path.exists():
        logger.warning("File %s not found.", path)
        return None

    if path.suffix == ".sdf":
        pass
    elif path.suffix in {".mol", ".mol2", ".pdb"}:
        yield safe_load_mol(path, sanitize=sanitize, **load_params)
        return None
    else:
        raise ValueError(f"Molecule file {path} has unknown format. Only .sdf, .mol, .mol2, and .pdb are supported.")

    with SDMolSupplier(str(path), sanitize=False, strictParsing=True) as supplier:
        # do not sanitize in supplier, this is done in process_mol and yields None for failed mols
        for index in range(len(supplier)) if indices is None else indices:
            mol_clean = _process_mol(supplier[index], sanitize=sanitize, **load_params)
            if mol_clean is not None:
                mol_clean.SetProp("_Path", str(path))
                mol_clean.SetProp("_Index", str(index))
            yield mol_clean


def get_num_mols(path: Path | Mol) -> int:
    """Get number of molecules in a file. Only supports multiple molecules in SDF files."""

    if isinstance(path, Mol):
        return 1

    path = Path(path)
    if not path.exists():
        return 0
    if path.suffix == ".sdf":
        with SDMolSupplier(str(path), sanitize=False, strictParsing=False) as supplier:
            return len(supplier)
    return 1


def _load_mol(  # noqa: PLR0913
    path: Path,
    load_all=False,
    sanitize=False,
    removeHs=False,
    strictParsing=False,
    proximityBonding=False,
    cleanupSubstructures=True,
    index: int | None = None,
    **params,
) -> Mol | None:
    """Load molecule(s) from a file, picking the right RDKit function."""

    if path.suffix == ".sdf":
        if load_all:
            mol = _load_and_combine_mols(path, sanitize=False, removeHs=removeHs, strictParsing=strictParsing)
        elif index is not None:
            with SDMolSupplier(str(path), sanitize=False, removeHs=removeHs, strictParsing=strictParsing) as supplier:
                mol = supplier[index]
        else:
            mol = MolFromMolFile(str(path), sanitize=False, removeHs=removeHs, strictParsing=strictParsing)

    elif path.suffix == ".mol2":
        # MolFromMol2File only loads only first molecule from mol2 file
        with open(path) as file:
            if load_all and sum(ln.strip().startswith("@<TRIPOS>MOLECULE") for ln in file.readlines()) > 1:
                logger.error("Cannot load multiple molecules from mol2 file, only loading first.")
        mol = MolFromMol2File(str(path), sanitize=False, removeHs=removeHs, cleanupSubstructures=cleanupSubstructures)

    elif path.suffix == ".pdb":
        mol = MolFromPDBFile(str(path), sanitize=False, removeHs=removeHs, proximityBonding=proximityBonding)

    elif path.suffix == ".mol":
        # .mol files only contain one molecule
        with open(path) as file:
            block = "".join(file.readlines()).strip() + "\nM  END"
            mol = MolFromMolBlock(block, sanitize=False, removeHs=removeHs, strictParsing=strictParsing)

    else:
        raise ValueError(f"Unknown file type {path.suffix}")

    if mol is not None:
        mol.SetProp("_Path", str(path))

    mol = _process_mol(mol, sanitize=sanitize, **params)

    return mol


def _load_and_combine_mols(path: Path, sanitize=True, removeHs=True, strictParsing=True) -> Mol | None:
    """Load mols from SDF file and combine into one molecule."""
    with SDMolSupplier(str(path), sanitize=sanitize, removeHs=removeHs, strictParsing=strictParsing) as supplier:
        # warning: combines molecules without checking identity or atom order
        mol = next(supplier)
        while mol is None and supplier.atEnd() is False:
            mol = next(supplier)
        for mol_next in supplier:
            if mol_next is not None:
                mol.AddConformer(mol_next.GetConformer(), assignId=True)
        return mol
    return None


def _process_mol(  # noqa: PLR0913
    mol: Mol | None,
    smiles: str | None = None,
    cleanup=False,
    sanitize=True,
    add_hs=False,
    assign_stereo=True,
) -> Mol:
    """Process a molecule, optionally adding hydrogens and assigning bond orders."""
    if mol is None:
        raise ValueError("Could not load molecule.")
    if smiles is not None:
        mol = _assign_bond_order(mol, smiles)
    if cleanup:
        mol = _cleanup(mol)
    if sanitize:
        mol = _sanitize(mol)
    if add_hs:
        mol = _add_hydrogens(mol)
    if assign_stereo:
        mol = _assign_stereo(mol)
    return mol


def _assign_bond_order(mol: Mol, smiles) -> Mol:
    template = MolFromSmiles(smiles)
    mol = AssignBondOrdersFromTemplate(template, mol)
    if mol is None:
        raise ValueError("Could not assign bond orders to molecule.")
    return mol


def _cleanup(mol: Mol) -> Mol:
    Cleanup(mol)
    if mol is None:
        raise ValueError("Could not cleanup molecule.")
    return mol


def _sanitize(mol: Mol) -> Mol:
    flags = SanitizeMol(mol)
    if mol is None or flags != 0:
        raise ValueError("Could not sanitize molecule.")
    return mol


def _add_hydrogens(mol: Mol) -> Mol:
    mol = AddHs(mol, addCoords=True)
    if mol is None:
        raise ValueError("Could not add hydrogens to molecule.")
    return mol


def _assign_stereo(mol: Mol) -> Mol:
    AssignStereochemistryFrom3D(mol)
    if mol is None:
        raise ValueError("Could not assign stereo to molecule.")
    return mol



"""Provides functions for calculating errors."""

from __future__ import annotations


def ae(val_pred: float, val_true: float) -> float:
    """Calculate absolute error."""
    return abs(val_pred - val_true)


def pe(val_pred: float, val_true: float) -> float:
    """Calculate percentage error."""
    return (val_pred - val_true) / val_true


def ape(val_pred: float, val_true: float) -> float:
    """Calculate absolute percentage error."""
    return abs(pe(val_pred, val_true))


def bae(val: float, lb: float, ub: float) -> float:
    """Calculate out of bounds absolute error."""
    if val < lb:
        return ae(val, lb)
    if val > ub:
        return ae(val, ub)
    return 0.0


def bpe(val: float, lb: float, ub: float) -> float:
    """Calculate out of bounds percentage error."""
    if val < lb:
        return pe(val, lb)
    if val > ub:
        return pe(val, ub)
    return 0.0


def bape(val: float, lb: float, ub: float) -> float:
    """Calculate out of bounds absolute percentage error."""
    if val < lb:
        return ape(val, lb)
    if val > ub:
        return ape(val, ub)
    return 0.0



"""PoseBusters class for running all tests on a set of molecules."""

from __future__ import annotations

import inspect
import logging
from collections.abc import Callable, Generator, Iterable
from concurrent.futures import ProcessPoolExecutor, as_completed
from concurrent.futures.process import BrokenProcessPool
from functools import partial
from math import ceil
from pathlib import Path
from typing import Any

import pandas as pd
from rdkit.Chem.rdchem import Mol
from yaml import safe_load

from .modules.distance_geometry import check_geometry
from .modules.energy_ratio import check_energy_ratio
from .modules.flatness import check_flatness
from .modules.identity import check_identity
from .modules.intermolecular_distance import check_intermolecular_distance
from .modules.loading import check_loading
from .modules.rmsd import check_rmsd
from .modules.sanity import (
    check_all_atoms_connected,
    check_chemistry,
    check_chemistry_using_inchi,
    check_chemistry_using_rdkit,
    check_radicals,
)
from .modules.sucos import check_sucos
from .modules.volume_overlap import check_volume_overlap
from .tools.loading import get_num_mols, safe_load_mol, safe_supply_mols

logger = logging.getLogger(__name__)


module_dict: dict[str, Callable] = {
    "loading": check_loading,
    "sanity": check_chemistry,  # for backwards compatibility
    "rdkit_sanity": check_chemistry_using_rdkit,
    "inchi_convertible": check_chemistry_using_inchi,
    "atoms_connected": check_all_atoms_connected,
    "check_radicals": check_radicals,
    "identity": check_identity,
    "distance_geometry": check_geometry,
    "flatness": check_flatness,
    "energy_ratio": check_energy_ratio,
    "intermolecular_distance": check_intermolecular_distance,
    "volume_overlap": check_volume_overlap,
    "rmsd": check_rmsd,
    "sucos": check_sucos,
}
molecule_args = {"mol_cond", "mol_true", "mol_pred"}

ResultKey = tuple[str, str, int]
ResultList = list[tuple[str, str, Any]]
ResultTuple = tuple[ResultKey, ResultList]
ResultDict = dict[ResultKey, ResultList]


class PoseBusters:
    """Class to run all tests on a set of molecules."""

    file_paths: pd.DataFrame
    module_name: list[str]
    module_func: list[Callable]
    module_args: list[set[str]]
    fname: list[str]

    def __init__(
        self,
        config: str | dict[str, Any] = "redock",
        top_n: int | None = None,
        max_workers: int | None = 0,
        chunk_size: int | None = 100,
    ) -> None:
        """Initialize PoseBusters object.

        Args:
            config: Configuration file or dictionary. If a string, it should be one of "dock", "redock", "mol", "gen".
            top_n: Number of poses to process. If None, all poses are processed.
            max_workers: Maximum number of workers for parallelization. If None, all available cores are used. If 0 or
                negative, no parallelization is used.
            chunk_size: Number of poses to process per process if parallelization is used. If None, parallelization over
                files only.
        """

        if isinstance(config, str) and config in {
            "dock",
            "redock",
            "mol",
            "gen",
            "regen",
            "dock_fast",
            "redock_fast",
            "mol_fast",
            "gen_fast",
            "regen_fast",
        }:
            logger.info("Using default configuration for mode %s.", config)
            with open(Path(__file__).parent / "config" / f"{config}.yml", encoding="utf-8") as config_file:
                self.config = safe_load(config_file)
        elif isinstance(config, dict):
            logger.info("Using configuration dictionary provided by user.")
            self.config = config
        else:
            logger.error("Configuration %s not valid. Provide 'dock', 'redock', 'mol', 'gen' or a dictionary.", config)

        # Validate that all function names in modules exist in module_dict
        if "modules" in self.config:
            invalid_functions = []
            for module in self.config["modules"]:
                func_name = module.get("function")
                if func_name and func_name not in module_dict:
                    invalid_functions.append(func_name)
            if invalid_functions:
                raise ValueError(
                    f"Invalid function names in config: {invalid_functions}. "
                    f"Available functions: {sorted(module_dict.keys())}"
                )

        self.config["top_n"] = self.config.get("top_n", top_n)
        self.config["max_workers"] = self.config.get("max_workers", max_workers)
        self.config["chunk_size"] = self.config.get("chunk_size", chunk_size)

    def bust(
        self,
        mol_pred: Iterable[Mol | Path | str] | Mol | Path | str,
        mol_true: Mol | Path | str | None = None,
        mol_cond: Mol | Path | str | None = None,
        full_report: bool = False,
    ) -> pd.DataFrame:
        """Run tests on one or more molecules.

        Args:
            mol_pred: Generated molecule(s), e.g. de-novo generated molecule or docked ligand, with one or more poses.
            mol_true: True molecule, e.g. crystal ligand, with one or more poses.
            mol_cond: Conditioning molecule, e.g. protein.
            full_report: Whether to include all columns in the output or only the boolean ones specified in the config.

        Notes:
            - Molecules can be provided as rdkit molecule objects or file paths.

        Returns:
            Pandas dataframe with results.
        """
        mol_pred_list: Iterable[Mol | Path | str] = [mol_pred] if isinstance(mol_pred, (Mol, Path, str)) else mol_pred

        columns = ["mol_pred", "mol_true", "mol_cond"]
        self.file_paths = pd.DataFrame([[mol_pred, mol_true, mol_cond] for mol_pred in mol_pred_list], columns=columns)
        generator = self._run()
        results = self._collect_in_table(generator, full_report=full_report)
        return results

    def bust_table(self, mol_table: pd.DataFrame, full_report: bool = False) -> pd.DataFrame:
        """Run tests on molecules provided in pandas dataframe as paths or rdkit molecule objects.

        Args:
            mol_table: Pandas dataframe with columns "mol_pred", "mol_true", "mol_cond" containing paths to molecules.
            full_report: Whether to include all columns in the output or only the boolean ones specified in the config.

        Returns:
            Pandas dataframe with results.
        """
        self.file_paths = mol_table
        generator = self._run()
        results = self._collect_in_table(generator, full_report=full_report)
        return results

    def _run(self) -> Generator[ResultTuple]:
        """Run all tests on molecules provided in file paths.

        Yields:
            Generator of result dictionaries.
        """
        self._initialize_modules()
        max_workers = self.config.get("max_workers", None)
        chunk_size = self.config.get("chunk_size", 100)
        if max_workers is not None and max_workers <= 0:
            yield from self._run_single_thread()
        elif chunk_size is None:
            yield from self._run_parallel_over_files(max_workers=max_workers)
        else:
            yield from self._run_parallel_over_poses(max_workers=max_workers, chunk_size=chunk_size)

    def _run_single_thread(self) -> Generator[ResultTuple]:
        for _, paths in self.file_paths.iterrows():
            yield from self._run_multiple_poses(paths)

    def _run_parallel_over_files(
        self, timeout: int | None = None, max_workers: int | None = None
    ) -> Generator[ResultTuple]:
        with ProcessPoolExecutor(max_workers=max_workers) as executor:
            futures = [executor.submit(self._run_and_combine, paths) for _, paths in self.file_paths.iterrows()]
            for future in as_completed(futures, timeout=None):
                try:
                    results = future.result(timeout=timeout)
                except BrokenProcessPool as exception:
                    # logger.critical("BrokenProcessPool: %s", exception)
                    raise exception
                except Exception as exception:
                    # logger.critical("Error in process: %s", exception)
                    raise exception

                yield from results

    def _run_parallel_over_poses(
        self, timeout: int | None = None, max_workers: int | None = None, chunk_size: int = 100
    ) -> Generator[ResultTuple]:
        with ProcessPoolExecutor(max_workers=max_workers) as executor:
            futures = []
            for _, paths in self.file_paths.iterrows():
                num_mols_pred = get_num_mols(paths["mol_pred"])
                for chunk in range(ceil(num_mols_pred / chunk_size)):
                    indices = range(chunk * chunk_size, min((chunk + 1) * chunk_size, num_mols_pred))
                    future = executor.submit(self._run_and_combine, paths=paths, indices=indices)
                    futures.append(future)

            for future in as_completed(futures, timeout=None):
                try:
                    results = future.result(timeout=timeout)
                except BrokenProcessPool as exception:
                    # logger.critical("BrokenProcessPool: %s", exception)
                    raise exception
                except Exception as exception:
                    # logger.critical("Error in process: %s", exception)
                    raise exception

                yield from results

    def _run_and_combine(self, paths: pd.Series, indices: Iterable[int] | None = None) -> list[ResultTuple]:
        """Run and collect all tests for all poses in the prediction file."""
        return list(self._run_multiple_poses(paths, indices=indices))

    def _run_multiple_poses(self, paths: pd.Series, indices: Iterable[int] | None = None) -> Generator[ResultTuple]:
        """Run all tests on indexed poses in the prediction file.

        Args:
            paths: Pandas series with keys "mol_pred", "mol_true", "mol_cond" containing paths to molecules.
            indices: Indices of poses to process. If None, all poses are processed.

        Yields:
            Generator of result dictionaries.
        """

        mol_args = {}
        if "mol_cond" in paths and paths["mol_cond"] is not None and pd.notna(paths["mol_cond"]):
            mol_cond_load_params = self.config.get("loading", {}).get("mol_cond", {})
            mol_args["mol_cond"] = safe_load_mol(path=paths["mol_cond"], **mol_cond_load_params)
        if "mol_true" in paths and paths["mol_true"] is not None and pd.notna(paths["mol_true"]):
            mol_true_load_params = self.config.get("loading", {}).get("mol_true", {})
            mol_args["mol_true"] = safe_load_mol(path=paths["mol_true"], **mol_true_load_params)

        mol_pred_load_params = self.config.get("loading", {}).get("mol_pred", {})
        for i, mol_pred in enumerate(safe_supply_mols(paths["mol_pred"], indices=indices, **mol_pred_load_params)):
            if self.config["top_n"] is not None and i >= self.config["top_n"]:
                break
            mol_args["mol_pred"] = mol_pred

            key: ResultKey = (str(paths["mol_pred"]), self._get_name(mol_pred), i)
            results: ResultList = self._run_one_pose(mol_args)

            yield key, results

    def _run_one_pose(self, molecules: dict[str, Any]) -> ResultList:
        """Run all tests on a single pose."""
        results = []
        for name, fname, func, args in zip(self.module_name, self.fname, self.module_func, self.module_args):
            # pick needed arguments for module
            args_needed = {k: v for k, v in molecules.items() if k in args}

            # loading takes all inputs
            if fname == "loading":
                args_needed = {k: args_needed.get(k, None) for k in args_needed}

            # run module when all needed input molecules are valid Mol objects
            if fname != "loading" and not all(args_needed.get(m, None) for m in args_needed):
                module_output: dict[str, Any] = {"results": {}}
            else:
                module_output = func(**args_needed)

            # save to object
            results.extend([(name, k, v) for k, v in module_output["results"].items()])
            # self.results[results_key]["details"].append(module_output["details"])

        return results

    def _initialize_modules(self) -> None:
        self.module_name = []
        self.module_func = []
        self.module_args = []
        self.fname = []
        for module in self.config["modules"]:
            function = module_dict[module["function"]]
            parameters = module.get("parameters", {})
            module_args = set(inspect.signature(function).parameters).intersection({"mol_pred", "mol_true", "mol_cond"})

            self.module_name.append(module["name"])
            self.fname.append(module["function"])
            self.module_func.append(partial(function, **parameters))
            self.module_args.append(module_args)

    @staticmethod
    def _get_name(mol: Mol | None) -> str:
        """Get the name of a molecule from the RDKit molecule object. Returns empty string if no name found."""
        if mol is None or not mol.HasProp("_Name"):
            return ""
        return str(mol.GetProp("_Name", autoConvert=False))

    def _collect_in_table(self, results_gen: Generator, full_report: bool) -> pd.DataFrame:
        """Collect generator results in a pandas dataframe."""

        df = pd.concat([self._make_table({k: v}, self.config, full_report=full_report) for k, v in results_gen])
        df.index.names = ["file", "molecule", "position"]
        df.columns = [c.lower().replace(" ", "_") for c in df.columns]

        return df

    @staticmethod
    def _make_table(results_dict: ResultDict, config: dict[str, Any], full_report: bool = False) -> pd.DataFrame:
        """Generate a table from the output of the tests."""

        d = {id: {(module, output): value for module, output, value in results} for id, results in results_dict.items()}
        df = pd.DataFrame.from_dict(d, orient="index")

        test_columns = [(c["name"], n) for c in config["modules"] for n in c.get("chosen_binary_test_output", [])]
        names_lookup = {(c["name"], k): v for c in config["modules"] for k, v in c.get("rename_outputs", {}).items()}
        suffix_lookup = {c["name"]: c["rename_suffix"] for c in config["modules"] if "rename_suffix" in c}

        available_columns = df.columns.tolist()
        missing_columns = [c for c in test_columns if c not in available_columns]
        extra_columns = [c for c in available_columns if c not in test_columns]
        columns = test_columns + extra_columns if full_report else test_columns

        df[missing_columns] = pd.NA
        df = df[columns]
        df.columns = [names_lookup.get(c, c[-1] + suffix_lookup.get(c[0], "")) for c in df.columns]  # type: ignore[call-overload]

        return df



from __future__ import annotations

from pathlib import Path

from posebusters.cli import _parse_args, bust

mols_table = "tests/conftest/sample_bust_docks_table.csv"
mol_pred_1ia1 = "tests/conftest/1IA1_TQ3/1IA1_TQ3_ligand.sdf"
mol_true_1ia1 = "tests/conftest/1IA1_TQ3/1IA1_TQ3_ligands.sdf"
mol_cond_1ia1 = "tests/conftest/1IA1_TQ3/1IA1_TQ3_no_ligand_no_solvent.pdb"


def test_parse_args() -> None:
    args = _parse_args(
        [
            mol_pred_1ia1,
            "-l",
            mol_true_1ia1,
            "-p",
            mol_cond_1ia1,
            "--full-report",
            "--outfmt",
            "csv",
        ]
    )
    assert args.mol_pred[0] == Path(mol_pred_1ia1)
    assert args.mol_true == Path(mol_true_1ia1)
    assert args.mol_cond == Path(mol_cond_1ia1)
    assert args.full_report
    assert args.outfmt == "csv"

    args = _parse_args([mol_pred_1ia1, "--outfmt", "long"])
    assert args.mol_pred[0] == Path(mol_pred_1ia1)
    assert not args.full_report
    assert args.outfmt == "long"


def test_bust_mols() -> None:
    bust([Path(mol_pred_1ia1)], Path(mol_true_1ia1), Path(mol_cond_1ia1))


def test_bust_table() -> None:
    bust(table=Path(mols_table))


def test_output() -> None:
    output = Path(".test_output.csv")
    bust(table=Path(mols_table), outfmt="csv", output=output)
    assert output.exists()


def test_bust_none() -> None:
    # check that ValueError is raised when no molecules are provided
    try:
        bust()
    except ValueError:
        pass
    else:
        raise AssertionError("ValueError not raised.")



"""Protein related functions."""

from __future__ import annotations

import logging
from collections.abc import Iterable

from rdkit.Chem.rdchem import Atom, Mol

logger = logging.getLogger(__name__)
_inorganic_cofactor_elements = {
    "Li",
    "Be",
    "Na",
    "Mg",
    "Cl",
    "K",
    "Ca",
    "Mn",
    "Fe",
    "Co",
    "Ni",
    "Cu",
    "Zn",
    "Br",
    "Rb",
    "Mo",
    "Cd",
}
_inorganic_cofactor_ccd_codes = {
    "FES",
    "MOS",
    "PO3",
    "PO4",
    "PPK",
    "SO3",
    "SO4",
    "VO4",
}
_water_ccd_codes = {
    "HOH",
}


def get_atom_type_mask(mol: Mol, ignore_types: Iterable[str]) -> list[bool]:
    """Get mask for atoms to keep."""
    ignore_types = set(ignore_types)
    unsupported = ignore_types - {"hydrogens", "protein", "organic_cofactors", "inorganic_cofactors", "waters"}
    if unsupported:
        raise ValueError(f"Ignore types {unsupported} not supported.")

    if mol.GetAtomWithIdx(0).GetPDBResidueInfo() is None:
        logger.warning("No PDB information found. Assuming organic molecule.")

    ignore_h = "hydrogens" in ignore_types
    ignore_protein = "protein" in ignore_types
    ignore_org_cof = "organic_cofactors" in ignore_types
    ignore_inorg_cof = "inorganic_cofactors" in ignore_types
    ignore_water = "waters" in ignore_types

    return [
        _keep_atom(a, ignore_h, ignore_protein, ignore_org_cof, ignore_inorg_cof, ignore_water) for a in mol.GetAtoms()
    ]


def _keep_atom(  # noqa: PLR0913, PLR0911
    atom: Atom, ignore_h: bool, ignore_protein: bool, ignore_org_cof: bool, ignore_inorg_cof: bool, ignore_water: bool
) -> bool:
    """Whether to keep atom for given ignore flags."""
    symbol = atom.GetSymbol()
    if ignore_h and symbol == "H":
        return False

    is_inorganic_elem = symbol in _inorganic_cofactor_elements
    if ignore_inorg_cof and is_inorganic_elem:
        return False

    # if loaded from PDB file, we can use the residue names and the hetero flag
    info = atom.GetPDBResidueInfo()
    if info is None:
        if ignore_org_cof:
            return False
        return True

    is_hetero = info.GetIsHeteroAtom()
    if ignore_protein and not is_hetero:
        return False

    residue_name = info.GetResidueName()
    is_water = residue_name in _water_ccd_codes
    if ignore_water and is_water:
        return False

    is_inorganic_ccd = residue_name in _inorganic_cofactor_ccd_codes
    if ignore_inorg_cof and is_inorganic_ccd:
        return False

    if ignore_org_cof and is_hetero and (not is_water) and (not is_inorganic_ccd) and (not is_inorganic_elem):
        return False

    return True



from __future__ import annotations

from posebusters.modules.flatness import check_flatness, flat_bonds, flat_rings, nonflat


def test_check_flatness(mol_pm2, mol_cgb):
    # no rings
    out = check_flatness(mol_cgb)
    assert out["results"]["flatness_passes"] is True

    # rings
    out = check_flatness(mol_pm2)
    assert out["results"]["flatness_passes"] is True


def test_check_flatness_6yr2_t1c(mol_pred_6yr2_t1c, mol_true_6yr2_t1c):
    # pass
    out = check_flatness(mol_true_6yr2_t1c, threshold_flatness=0.5)
    assert out["results"]["flatness_passes"] is True

    # fail
    out = check_flatness(mol_pred_6yr2_t1c)
    assert out["results"]["flatness_passes"] is False

    # fail rings
    out = check_flatness(mol_pred_6yr2_t1c, flat_systems=flat_rings)
    assert out["results"]["flatness_passes"] is False

    # pass bonds
    out = check_flatness(mol_pred_6yr2_t1c, flat_systems=flat_bonds, threshold_flatness=0.5)
    assert out["results"]["flatness_passes"] is False


def test_check_flatness_7ecr_sin(mol_lig_7ecr_sin):
    # pass because no rings or double bonds present
    out = check_flatness(mol_lig_7ecr_sin)
    assert out["results"]["flatness_passes"] is True


def test_check_nonflatness_synthetic(mols_flat_etkdgv3, mols_nonflat_etkdgv3):
    for mol in mols_flat_etkdgv3:
        # these are flat and should be ignored by the patterns
        out = check_flatness(mol, flat_systems=nonflat, check_nonflat=True)
        assert out["results"]["flatness_passes"]

    for mol in mols_nonflat_etkdgv3:
        # these should be non-flat
        out = check_flatness(mol, flat_systems=nonflat, check_nonflat=True)
        assert out["results"]["flatness_passes"]


def test_check_nonflatness_awj_ideal(mol_awj_ideal):
    # RDKit does not recognize the 5-membered ring as aromatic (2 < 4n + 2 for all n)
    out = check_flatness(mol_awj_ideal)
    assert out["results"]["flatness_passes"] is True
    out = check_flatness(mol_awj_ideal, flat_systems=nonflat, check_nonflat=True, threshold_flatness=0.05)
    assert out["results"]["flatness_passes"] is True


def test_check_nonflatness_o2u_ideal(mol_o2u_ideal):
    out = check_flatness(mol_o2u_ideal)
    assert out["results"]["flatness_passes"] is True
    out = check_flatness(mol_o2u_ideal, flat_systems=nonflat, check_nonflat=True, threshold_flatness=0.05)
    assert out["results"]["flatness_passes"] is True


def test_check_nonflatness_pip_ideal(mol_pip_ideal):
    out = check_flatness(mol_pip_ideal)
    assert out["results"]["flatness_passes"] is True
    out = check_flatness(mol_pip_ideal, flat_systems=nonflat, check_nonflat=True, threshold_flatness=0.05)
    assert out["results"]["flatness_passes"] is True


def test_check_nonflatness_pip_wrong(mol_pip_wrong):
    out = check_flatness(mol_pip_wrong)
    assert out["results"]["flatness_passes"] is True
    out = check_flatness(mol_pip_wrong, flat_systems=nonflat, check_nonflat=True, threshold_flatness=0.05)
    assert out["results"]["flatness_passes"] is False



"""Module to check energy of ligand conformations."""

from __future__ import annotations

import logging
from copy import deepcopy
from functools import cache
from math import isfinite

from rdkit import ForceField  # noqa: F401
from rdkit.Chem.inchi import InchiReadWriteError, MolFromInchi
from rdkit.Chem.rdchem import Mol
from rdkit.Chem.rdDistGeom import EmbedMultipleConfs, ETKDGv3
from rdkit.Chem.rdForceFieldHelpers import (
    UFFGetMoleculeForceField,
    UFFHasAllMoleculeParams,
    UFFOptimizeMoleculeConfs,
)
from rdkit.Chem.rdmolops import AddHs, AssignStereochemistryFrom3D, SanitizeMol

from ..tools.inchi import get_inchi
from ..tools.logging import CaptureLogger
from ..tools.molecules import add_hydrogens_with_uff_positions

logger = logging.getLogger(__name__)


_empty_results = {
    "results": {
        "num_h_added": float("nan"),
        "mol_pred_energy": float("nan"),
        "ensemble_avg_energy": float("nan"),
        "energy_ratio": float("nan"),
        "energy_ratio_passes": float("nan"),
    }
}


def check_energy_ratio(
    mol_pred: Mol,
    threshold_energy_ratio: float = 7.0,
    ensemble_number_conformations: int = 100,
    inchi_strict: bool = False,
    epsilon=1e-10,
    num_threads=0,
) -> dict[str, dict[str, float | bool]]:
    """Check whether the internal energy of a molecular conformation is too far from its ground state.

    Notes:
      - If there are missing explicit hydrogens, they are added and their positions are optimized.
      - Hydrogens are missing when there are radicals or there are implicit hydrogens.
      - The energy ratio 'with hydrogens' uses the energy of the molecule after filling in hydrogens
        but before optimizing their positions.
      - The energy ratio 'without hydrogens' uses the energy of the molecule after filling in hydrogens
        AND after optimizing their positions. So their contribution (if any) to the energy ratio is reduced.

    Args:
        mol_pred: Predicted molecule (docked ligand) with exactly one conformer.
        threshold_energy_ratio: Limit above which the energy ratio is deemed to high. Defaults to 7.0.
        ensemble_number_conformations: Number of conformations to generate for the ensemble over which to
            average. Defaults to 100.
        inchi_strict: Whether to treat warnings in the InChI generation as errors. Defaults to False.
        num_threads: The number of threads to use for energy minimization.
            By default, the number of available cores is used.

    Returns:
        PoseBusters results dictionary.
    """
    mol_pred = deepcopy(mol_pred)

    try:
        assert mol_pred.GetNumConformers() > 0, "Molecule does not have a conformer."
        assert not SanitizeMol(mol_pred, catchErrors=True), "Molecule does not sanitize."
    except Exception as e:
        logger.warning("Energy ratio module failed because %s", e)
        return _empty_results

    try:
        with CaptureLogger():
            num_atoms_in = mol_pred.GetNumAtoms()
            # make hydrogens explicit, replace radicals with hydrogens, optimize added hydrogens' positions
            mol_pred, _, observed_energy = add_hydrogens_with_uff_positions(mol_pred)
            num_atoms_filled = mol_pred.GetNumAtoms()
            num_h_added = num_atoms_filled - num_atoms_in
        assert UFFHasAllMoleculeParams(mol_pred), "UFF parameters missing for molecule."
    except Exception as e:
        logger.warning("Energy ratio module failed because %s", e)
        return _empty_results

    try:
        inchi = get_inchi(mol_pred, inchi_strict=inchi_strict)
    except InchiReadWriteError as e:
        logger.warning("Energy ratio module failed because InChI creation failed for molecule: %s", e)
        return _empty_results
    except Exception as e:
        logger.warning("Energy ratio module failed because InChI creation failed for molecule: %s", e)
        return _empty_results

    try:
        energies = get_energies(inchi, ensemble_number_conformations, num_threads)
        mean_energy = sum(energies) / len(energies)
    except Exception as e:
        logger.warning("Energy ratio module failed to calculate ensemble conformation energy for %s: %s", inchi, e)
        mean_energy = float("nan")

    if mean_energy == 0:
        logger.warning("Energy ratio module calculated an average energy of 0 for %s", inchi)
        mean_energy = epsilon  # clipping

    energy_ratio = observed_energy / mean_energy
    energy_ratio_passes = energy_ratio <= threshold_energy_ratio if isfinite(energy_ratio) else float("nan")

    results = {
        "num_h_added": num_h_added,
        "mol_pred_energy": observed_energy,
        "ensemble_avg_energy": mean_energy,
        "energy_ratio": energy_ratio,
        "energy_ratio_passes": energy_ratio_passes,
    }
    return {"results": results}


def get_average_energy(inchi: str, n_confs: int = 50, num_threads: int = 0) -> float:
    """Get average energy of an ensemble of molecule conformations."""
    energies = get_energies(inchi, n_confs, num_threads)
    return float(sum(energies) / len(energies))


@cache
def get_energies(inchi: str, n_confs: int = 50, num_threads: int = 0) -> list[float]:
    """Get energies of an ensemble of molecule conformations."""
    with CaptureLogger():
        mol = MolFromInchi(inchi)
    return new_conformation(mol, n_confs, num_threads)["energies"]


def new_conformation(
    mol: Mol, n_confs: int = 1, num_threads: int = 0, energy_minimization=True
) -> dict[str, Mol | list[float]]:
    """Generate new conformation(s) for a molecule."""
    assert mol is not None

    etkdg = ETKDGv3()
    etkdg.randomSeed = 42  # type: ignore[assignment]
    etkdg.verbose = False  # type: ignore[assignment]
    etkdg.useRandomCoords = True  # type: ignore[assignment]
    etkdg.numThreads = num_threads  # type: ignore[assignment]

    # prep mol
    mol_etkdg = deepcopy(mol)
    mol_etkdg = AddHs(mol_etkdg, addCoords=True)
    AssignStereochemistryFrom3D(mol_etkdg, replaceExistingTags=True)
    mol_etkdg.RemoveAllConformers()

    # etkdg
    with CaptureLogger():
        cids_etkdg = EmbedMultipleConfs(mol_etkdg, n_confs, etkdg)
        assert len(cids_etkdg) == n_confs, "Failed to generate conformations."

    # energy minimization
    if energy_minimization:
        with CaptureLogger():
            energy_uff = UFFOptimizeMoleculeConfs(mol_etkdg, numThreads=num_threads)
        energy_uff = [v[1] for v in energy_uff]
    else:
        energy_uff = [get_conf_energy(mol_etkdg, conf_id) for conf_id in cids_etkdg]

    return {"mol": mol_etkdg, "energies": energy_uff}


def get_conf_energy(mol: Mol, conf_id: int = -1) -> float:
    """Get energy of a conformation."""
    assert mol is not None

    mol = AddHs(mol, addCoords=True)
    with CaptureLogger():
        uff = UFFGetMoleculeForceField(mol, confId=conf_id)
        e_uff = uff.CalcEnergy()
    return e_uff



from __future__ import annotations

import numpy as np

from posebusters.modules.volume_overlap import check_volume_overlap


def test_check_volume_overlap(mol_rq3_x00, mol_rq3_x01, mol_rq3_x10):
    out = check_volume_overlap(mol_rq3_x00, mol_rq3_x00)
    assert out["results"]["volume_overlap"] == 1.0
    assert out["results"]["no_volume_clash"] is False

    out = check_volume_overlap(mol_rq3_x00, mol_rq3_x01)
    assert out["results"]["no_volume_clash"] is False

    out = check_volume_overlap(mol_rq3_x00, mol_rq3_x10)
    assert np.isnan(out["results"]["volume_overlap"])
    assert out["results"]["no_volume_clash"] is True

    out = check_volume_overlap(mol_rq3_x00, mol_rq3_x10, search_distance=20.0)
    assert out["results"]["volume_overlap"] == 0.0
    assert out["results"]["no_volume_clash"] is True


def test_check_volume_overlap_direction(mol_calcium, mol_cholesterol):
    out = check_volume_overlap(mol_calcium, mol_cholesterol, search_distance=100.0)
    assert out["results"]["volume_overlap"] > 0.8
    assert out["results"]["no_volume_clash"] is False

    out = check_volume_overlap(mol_cholesterol, mol_calcium, search_distance=100.0)
    assert out["results"]["volume_overlap"] < 0.2
    assert out["results"]["no_volume_clash"] is False


def test_check_volume_overlap_without_hydrogens(mol_rq3_x00, mol_rq3_x01, mol_rq3_x10):
    out = check_volume_overlap(mol_rq3_x00, mol_rq3_x00)
    assert out["results"]["volume_overlap"] == 1.0
    assert out["results"]["no_volume_clash"] is False

    out = check_volume_overlap(mol_rq3_x00, mol_rq3_x01)
    assert out["results"]["no_volume_clash"] is False

    out = check_volume_overlap(mol_rq3_x00, mol_rq3_x10)
    assert np.isnan(out["results"]["volume_overlap"])
    assert out["results"]["no_volume_clash"] is True

    out = check_volume_overlap(mol_rq3_x00, mol_rq3_x10, search_distance=20.0)
    assert out["results"]["volume_overlap"] == 0.0
    assert out["results"]["no_volume_clash"] is True


def test_check_volume_overlap_7cnq_g8x(mol_lig_7cnq_g8x, mol_cond_7cnq_g8x):
    out = check_volume_overlap(
        mol_lig_7cnq_g8x,
        mol_cond_7cnq_g8x,
        clash_cutoff=0.075,
        vdw_scale=0.8,
        ignore_types={"hydrogens", "inorganic_cofactors", "organic_cofactors"},
    )
    assert out["results"]["volume_overlap"] > 0.0
    assert out["results"]["no_volume_clash"] is True


def test_check_volume_overlap_7ztl_bcn(mol_lig_7ztl_bcn, mol_cond_7ztl_bcn):
    out = check_volume_overlap(
        mol_lig_7ztl_bcn,
        mol_cond_7ztl_bcn,
        clash_cutoff=0.075,
        vdw_scale=0.5,
        ignore_types={"hydrogens", "protein", "organic_cofactors"},
    )
    assert out["results"]["volume_overlap"] > 0.0
    assert out["results"]["no_volume_clash"] is True



"""Module to check RMSD between docked and crystal ligand."""

from __future__ import annotations

import logging
from copy import deepcopy

import numpy as np
from rdkit.Chem.AllChem import AssignBondOrdersFromTemplate
from rdkit.Chem.MolStandardize import rdMolStandardize
from rdkit.Chem.rdchem import Mol
from rdkit.Chem.rdMolAlign import CalcRMS, GetBestRMS
from rdkit.Chem.rdmolops import RemoveHs, RemoveStereochemistry
from rdkit.rdBase import LogToPythonLogger

from ..tools.logging import CaptureLogger
from ..tools.molecules import neutralize_atoms, remove_all_charges_and_hydrogens

LogToPythonLogger()
logger = logging.getLogger(__name__)

tautomer_enumerator = rdMolStandardize.TautomerEnumerator()
tautomer_enumerator.SetMaxTautomers(100000)
tautomer_enumerator.SetMaxTransforms(100000)
tautomer_enumerator.SetReassignStereo(True)
tautomer_enumerator.SetRemoveBondStereo(True)
tautomer_enumerator.SetRemoveSp3Stereo(True)


def check_rmsd(
    mol_pred: Mol, mol_true: Mol, rmsd_threshold: float = 2.0, heavy_only: bool = True, choose_by: str = "rmsd"
) -> dict[str, dict[str, bool | float]]:
    """Calculate RMSD and related metrics between predicted molecule and closest ground truth molecule.

    Args:
        mol_pred: Predicted molecule (docked ligand) with exactly one conformer.
        mol_true: Ground truth molecule (crystal ligand) with at least one conformer. If multiple conformers are
            present, the lowest RMSD will be reported.
        rmsd_threshold: Threshold in angstrom for reporting whether RMSD is within threshold. Defaults to 2.0.
        heavy_only: Whether to only consider heavy atoms for RMSD calculation. Defaults to True.
        choose_by: Metric to choose which mol_true conformation to compare to. Defaults to "rmsd".

    Returns:
        PoseBusters results dictionary.
    """
    assert isinstance(mol_true, Mol), "Ground truth molecule is missing."
    assert isinstance(mol_pred, Mol), "Predicted molecule is missing."
    num_conf = mol_true.GetNumConformers()
    assert num_conf > 0, "Crystal ligand needs at least one conformer."
    assert mol_pred.GetNumConformers() == 1, "Docked ligand should only have one conformer."

    rmsds = [robust_rmsd(mol_true, mol_pred, conf_id_probe=i, heavy_only=heavy_only) for i in range(num_conf)]
    kabsch_rmsds = [
        robust_rmsd(mol_true, mol_pred, conf_id_probe=i, kabsch=True, heavy_only=heavy_only) for i in range(num_conf)
    ]
    intercentroids = [
        intercentroid(mol_true, mol_pred, conf_id_probe=i, heavy_only=heavy_only) for i in range(num_conf)
    ]

    if choose_by == "rmsd":
        i = np.argmin(np.nan_to_num(rmsds, nan=np.inf))
    elif choose_by == "kabsch_rmsd":
        i = np.argmin(np.nan_to_num(kabsch_rmsds, nan=np.inf))
    elif choose_by == "centroid_distance":
        i = np.argmin(np.nan_to_num(intercentroids, nan=np.inf))
    else:
        raise ValueError(f"Invalid value {choose_by} for choose_by. Use 'rmsd', 'kabsch_rmsd', 'centroid_distance'.")
    rmsd = rmsds[i]
    kabsch_rmsd = kabsch_rmsds[i]
    centroid_dist = intercentroids[i]

    rmsd_within_threshold = rmsd <= rmsd_threshold

    results = {
        "rmsd": rmsd,
        "kabsch_rmsd": kabsch_rmsd,
        "centroid_distance": centroid_dist,
        "rmsd_within_threshold": rmsd_within_threshold,
    }
    return {"results": results}


def robust_rmsd(  # noqa: PLR0913
    mol_probe: Mol,
    mol_ref: Mol,
    conf_id_probe: int = -1,
    conf_id_ref: int = -1,
    drop_stereo: bool = False,
    heavy_only: bool = True,
    kabsch: bool = False,
    symmetrizeConjugatedTerminalGroups=True,
    **params,
) -> float:
    """RMSD calculation for isomers."""
    mol_probe = deepcopy(mol_probe)
    mol_ref = deepcopy(mol_ref)  # copy mols because rdkit RMSD calculation aligns mols

    if drop_stereo:
        RemoveStereochemistry(mol_probe)
        RemoveStereochemistry(mol_ref)

    if heavy_only:
        mol_probe = RemoveHs(mol_probe, sanitize=False)
        mol_ref = RemoveHs(mol_ref, sanitize=False)

    # combine parameters
    params = dict(symmetrizeConjugatedTerminalGroups=symmetrizeConjugatedTerminalGroups, kabsch=kabsch, **params)

    # calculate RMSD
    rmsd = _call_rdkit_rmsd(mol_probe, mol_ref, conf_id_probe, conf_id_ref, **params)
    if not np.isnan(rmsd):
        return rmsd

    # try again ignoring charges or tautomers
    rmsd = _rmsd_ignoring_charges_and_tautomers(mol_ref, mol_probe, conf_id_ref, conf_id_probe, params)
    if not np.isnan(rmsd):
        return rmsd

    # try assigning the bond orders of one molecule to the other
    try:
        mol_probe_new_bonds = AssignBondOrdersFromTemplate(mol_ref, mol_probe)
    except ValueError:
        return np.nan
    rmsd = _call_rdkit_rmsd(mol_probe_new_bonds, mol_ref, conf_id_probe, conf_id_ref, **params)
    if not np.isnan(rmsd):
        return rmsd

    return np.nan


def _rmsd_ignoring_charges_and_tautomers(
    mol_ref: Mol, mol_probe: Mol, conf_id_ref: int, conf_id_probe: int, params: dict
) -> float:
    # try again but remove charges and hydrogens
    mol_ref_uncharged = remove_all_charges_and_hydrogens(mol_ref)
    mol_probe_uncharged = remove_all_charges_and_hydrogens(mol_probe)
    rmsd = _call_rdkit_rmsd(mol_probe_uncharged, mol_ref_uncharged, conf_id_probe, conf_id_ref, **params)
    if not np.isnan(rmsd):
        return rmsd

    # try again but neutralize atoms
    mol_ref_neutralized = neutralize_atoms(mol_ref)
    mol_probe_neutralized = neutralize_atoms(mol_probe)
    rmsd = _call_rdkit_rmsd(mol_probe_neutralized, mol_ref_neutralized, conf_id_probe, conf_id_ref, **params)
    if not np.isnan(rmsd):
        return rmsd

    # try again but on canonical tautomers
    mol_ref_canonical = tautomer_enumerator.Canonicalize(mol_ref)
    mol_probe_canonical = tautomer_enumerator.Canonicalize(mol_probe)
    rmsd = _call_rdkit_rmsd(mol_probe_canonical, mol_ref_canonical, conf_id_probe, conf_id_ref, **params)
    if not np.isnan(rmsd):
        return rmsd

    # try again but after neutralizing atoms
    mol_ref_neutral_canonical = tautomer_enumerator.Canonicalize(neutralize_atoms(mol_ref))
    mol_probe_neutral_canonical = tautomer_enumerator.Canonicalize(neutralize_atoms(mol_probe))
    rmsd = _call_rdkit_rmsd(
        mol_probe_neutral_canonical, mol_ref_neutral_canonical, conf_id_probe, conf_id_ref, **params
    )
    if not np.isnan(rmsd):
        return rmsd

    return rmsd


def _call_rdkit_rmsd(mol_probe: Mol, mol_ref: Mol, conf_id_probe: int, conf_id_ref: int, **params) -> float:
    try:
        with CaptureLogger():
            return _rmsd(mol_probe, mol_ref, conf_id_probe, conf_id_ref, **params)
    except RuntimeError:
        pass
    except ValueError:
        pass

    return np.nan


def _rmsd(mol_probe: Mol, mol_ref: Mol, conf_id_probe: int, conf_id_ref: int, kabsch: bool = False, **params) -> float:
    if kabsch is True:
        return GetBestRMS(prbMol=mol_probe, refMol=mol_ref, prbId=conf_id_probe, refId=conf_id_ref, **params)
    return CalcRMS(prbMol=mol_probe, refMol=mol_ref, prbId=conf_id_probe, refId=conf_id_ref, **params)


def intercentroid(
    mol_probe: Mol, mol_ref: Mol, conf_id_probe: int = -1, conf_id_ref: int = -1, heavy_only: bool = True
) -> float:
    """Distance between centroids of two molecules."""

    centroid_probe = get_centroid(mol_probe, heavy_only, conf_id_probe)
    centroid_ref = get_centroid(mol_ref, heavy_only, conf_id_ref)
    return float(np.linalg.norm(centroid_probe - centroid_ref))


def get_centroid(mol: Mol, heavy_only: bool = True, conf_id: int = -1) -> np.ndarray:
    """Get centroid of molecule."""
    pos = mol.GetConformer(conf_id).GetPositions()
    if heavy_only:
        pos = pos[[atom.GetAtomicNum() != 1 for atom in mol.GetAtoms()], :]
    return pos.mean(axis=0)
